"""Finding a path where analysis files are saved"""

from dataclasses import dataclass
import inspect
import os


@dataclass
class InfoPath:
    """
    Path information for creating summaries
    """

    # Used in `sub_dir_name` to indicate that sub-directory should
    # not be created, and the file is placed in parent directory instead.
    DO_NOT_CREATE: str = "do not create"

    """
    Name of the top directory in which a sub-subdirectory `sub_dir_name`
    containing plots is placed.
    """
    dir_name: str = "model_info"

    """
    Name of the subdirectory. If None, determined automatically
    based on the name of the python script file from which a
    Tarpan's function is called.
    """
    sub_dir_name: str = None

    base_name: str = None  # Base name of the summary file or plot
    extension: str = None  # Extension of the summary or plot

    """
    DPI setting used for plots. Determines the quality of the plots
    """
    dpi: int = 300

    """
    Path to the python type from which a Tarpan function was called.
    This `codefile_path` variable is determined automatically.
    This path is then used to automatically create subdirectory
    `sub_dir_name` where plots are saved.
    """
    codefile_path: str = None

    """
    Full path to parent of info dir. Auto if None. If None, determined
    automatically based on the parent directory of the python script file
    from which a Tarpan's function is called.
    """
    path: str = None

    def set_codefile(self):
        """
        Sets the path to python file that called a Tarpan function.

        Returns
        --------
        InfoPath
            The path object with codefile_path set.
        """

        if self.codefile_path is not None:
            # The path is already set
            return

        tree_depth = 2
        frame = inspect.stack()[tree_depth]
        module = inspect.getmodule(frame[0])
        self.codefile_path = module.__file__

    def __str__(self):
        """
        Returns
        --------
        str:
            Get full path to the plot or summary file.
        """

        self.set_codefile()
        return get_info_path(self)

    def dir(self):
        """
        Returns
        --------
        str:
            Get full path to the directory where plot files are placed.
        """

        self.set_codefile()
        return get_info_dir(self)


def get_info_path(info_path=InfoPath()):
    """
    Get full path to the plot or summary file.

    Parameters
    ----------
    info_path : InfoPath
        Path information for creating summaries.


    Returns
    --------
    str
        Path to an analysis file.
    """

    info_path.set_codefile()
    full_path = get_info_dir(info_path)
    filename = f'{info_path.base_name}.{info_path.extension}'
    return os.path.join(full_path, filename)


def get_info_dir(info_path=InfoPath()):
    """
    Get full path to the directory where plots and summaries are placed

    Parameters
    ----------
    info_path : InfoPath
        Path information for creating summaries.


    Returns
    --------
    str
        Path to an analysis file.
    """

    info_path.set_codefile()

    if info_path.path is None:
        info_path.path = os.path.dirname(info_path.codefile_path)
    else:
        info_path.path = os.path.expanduser(info_path.path)

    full_path = os.path.join(info_path.path, info_path.dir_name)

    if info_path.sub_dir_name != info_path.DO_NOT_CREATE:
        if info_path.sub_dir_name is None:
            info_path.sub_dir_name = os.path.basename(info_path.codefile_path).rsplit('.', 1)[0]

        full_path = os.path.join(full_path, info_path.sub_dir_name)

    os.makedirs(full_path, exist_ok=True)

    return full_path



#
# The following code is taken from arviz library
# (https://github.com/arviz-devs/arviz), version 0.6.1,
# distributed under Apache License Version 2.0.
#
# Changes made to arviz code:
#  * Replaced _logsumexp with logsumexp from scipy.special
#
# Original code comes from Aki Vehtari, Tuomas Sivula:
#   https://github.com/avehtari/PSIS
#
# Theory: https://arxiv.org/abs/1507.02646v5
#

import numpy as np
from scipy.special import logsumexp


def _psislw(log_weights, cutoff_ind, cutoffmin, k_min=1.0 / 3):
    """
    Pareto smoothed importance sampling (PSIS) for a 1D vector.

    Parameters
    ----------
    log_weights : array
        Array of length n_observations
    cutoff_ind : int
    cutoffmin : float
    k_min : float

    Returns
    -------
    lw_out : array
        Smoothed log weights
    kss : float
        Pareto tail index
    """
    x = np.asarray(log_weights)

    # improve numerical accuracy
    x -= np.max(x)
    # sort the array
    x_sort_ind = np.argsort(x)
    # divide log weights into body and right tail
    xcutoff = max(x[x_sort_ind[cutoff_ind]], cutoffmin)

    expxcutoff = np.exp(xcutoff)
    (tailinds,) = np.where(x > xcutoff)  # pylint: disable=unbalanced-tuple-unpacking
    x_tail = x[tailinds]
    tail_len = len(x_tail)
    if tail_len <= 4:
        # not enough tail samples for gpdfit
        k = np.inf
    else:
        # order of tail samples
        x_tail_si = np.argsort(x_tail)
        # fit generalized Pareto distribution to the right tail samples
        x_tail = np.exp(x_tail) - expxcutoff
        k, sigma = _gpdfit(x_tail[x_tail_si])

        if k >= k_min:
            # no smoothing if short tail or GPD fit failed
            # compute ordered statistic for the fit
            sti = np.arange(0.5, tail_len) / tail_len
            smoothed_tail = _gpinv(sti, k, sigma)
            smoothed_tail = np.log(  # pylint: disable=assignment-from-no-return
                smoothed_tail + expxcutoff
            )
            # place the smoothed tail into the output array
            x[tailinds[x_tail_si]] = smoothed_tail
            # truncate smoothed values to the largest raw weight 0
            x[x > 0] = 0
    # renormalize weights
    x -= logsumexp(x)

    return x, k


def _gpdfit(ary):
    """Estimate the parameters for the Generalized Pareto Distribution (GPD).

    Empirical Bayes estimate for the parameters of the generalized Pareto
    distribution given the data.

    Parameters
    ----------
    ary : array
        sorted 1D data array

    Returns
    -------
    k : float
        estimated shape parameter
    sigma : float
        estimated scale parameter
    """
    prior_bs = 3
    prior_k = 10
    n = len(ary)
    m_est = 30 + int(n ** 0.5)

    b_ary = 1 - np.sqrt(m_est / (np.arange(1, m_est + 1, dtype=float) - 0.5))
    b_ary /= prior_bs * ary[int(n / 4 + 0.5) - 1]
    b_ary += 1 / ary[-1]

    k_ary = np.log1p(-b_ary[:, None] * ary).mean(axis=1)  # pylint: disable=no-member
    len_scale = n * (np.log(-(b_ary / k_ary)) - k_ary - 1)
    weights = 1 / np.exp(len_scale - len_scale[:, None]).sum(axis=1)

    # remove negligible weights
    real_idxs = weights >= 10 * np.finfo(float).eps
    if not np.all(real_idxs):
        weights = weights[real_idxs]
        b_ary = b_ary[real_idxs]
    # normalise weights
    weights /= weights.sum()

    # posterior mean for b
    b_post = np.sum(b_ary * weights)
    # estimate for k
    k_post = np.log1p(-b_post * ary).mean()  # pylint: disable=invalid-unary-operand-type,no-member
    # add prior for k_post
    k_post = (n * k_post + prior_k * 0.5) / (n + prior_k)
    sigma = -k_post / b_post

    return k_post, sigma


def _gpinv(probs, kappa, sigma):
    """Inverse Generalized Pareto distribution function."""
    # pylint: disable=unsupported-assignment-operation, invalid-unary-operand-type
    x = np.full_like(probs, np.nan)
    if sigma <= 0:
        return x
    ok = (probs > 0) & (probs < 1)
    if np.all(ok):
        if np.abs(kappa) < np.finfo(float).eps:
            x = -np.log1p(-probs)
        else:
            x = np.expm1(-kappa * np.log1p(-probs)) / kappa
        x *= sigma
    else:
        if np.abs(kappa) < np.finfo(float).eps:
            x[ok] = -np.log1p(-probs[ok])
        else:
            x[ok] = np.expm1(-kappa * np.log1p(-probs[ok])) / kappa
        x *= sigma
        x[probs == 0] = 0
        if kappa >= 0:
            x[probs == 1] = np.inf
        else:
            x[probs == 1] = -sigma / kappa
    return x



import shutil
import os
from tarpan.cmdstanpy.cache import run
from cmdstanpy import CmdStanModel


def myfunc(output_dir, data):
    data["count"] += 1

    data = {
        "J": 8,
        "y": [28,  8, -3,  7, -1,  1, 18, 12],
        "sigma": [15, 10, 16, 11,  9, 11, 10, 18]
    }

    model_path = "tarpan/testutils/a01_eight_schools/eight_schools.stan"
    model = CmdStanModel(stan_file=model_path)

    return model.sample(data=data, chains=1, cores=1,
                        sampling_iters=1000, warmup_iters=1000,
                        output_dir=output_dir)


def test_run():

    outdir = "tarpan/cmdstanpy/model_info/cache_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    data = {"count": 1}

    fit = run(myfunc, data=data)

    assert os.path.isfile(os.path.join(outdir, "stan_cache", "fit.pkl"))
    assert 'tau' in fit.column_names
    assert fit.get_drawset()['tau'].shape == (1000,)

    # Call it again
    fit = run(myfunc, data=data)

    assert 'tau' in fit.column_names
    assert fit.get_drawset()['tau'].shape == (1000,)

    # Ensure `myfunc` was called once
    assert data["count"] == 2



from tarpan.shared.compare import model_weights


def test_weights():
    weights = model_weights(deviances=[[1, 2, 3], [2, 3, 4], [7, 8, 9]])

    actual_weights = [
        round(weight, 5)
        for weight in weights
    ]

    assert actual_weights == [0.81749, 0.18241, 0.0001]



import numpy as np
from itertools import cycle
from dataclasses import dataclass, field
from typing import List
from tarpan.shared.info_path import InfoPath, get_info_path
from tarpan.plot.kde import scatter_and_kde, ScatterKdeParams


@dataclass
class PosteriorKdeParams:
    kde_plot_count: int = 50  # Number of posterior plots
    kde_values: int = 500  # Number of x and y values to calculate for KDE

    kde_edgecolors: List = field(
        default_factory=lambda:
            ["#0060ff15", '#ff002115', '#8888FF30', '#BBBB1130'])

    data_kde_show_shade: bool = False  # Show shade under data KDE plot
    data_kde_line_width: float = 2  # Line width of data KDE


def save_posterior_scatter_and_kde(
    fits,
    pdf,
    values,
    uncertainties,
    title=None,
    xlabel=None,
    ylabel=None,
    info_path=InfoPath(),
    scatter_kde_params=ScatterKdeParams(),
    posterior_kde_params=PosteriorKdeParams(),
    legend_labels=None,
    plot_fn=None):
    """
    Create a scatter plot for the data using `values` and `uncertainties`,
    show a KDE plot under it, and plot multiple posterior distributions
    on the KDE plot.

    Parameters
    ----------
    fits : list of cmdstanpy.stanfit.CmdStanMCMC
        Contains the samples from cmdstanpy.

    pdf : function(x, row)
        A function that will be called to evaluate the posterior
        distribution at x. The functinon should return the value
        of the posterior PDE corresponding to x.

        x : float
            Value at which the posterior is evaluated
        row : dict
            Contains single sample from posterior distribution.

    values: list of lists
        List of values to plot. Supply more than one list to
        see distributions shown with different colors and markers.
    uncertainties: list of lists
        Uncertainties coresponding to the `values`.
    plot_fn: [function(fig, axes, params), params]
        function:
            A function that can be used to add extra information to the
            plot before it is saved.

            Parameters
            ----------

            fig: Matplotlib's figure object
            axes: list of Matplotlib's axes objects
            params: custom parameters that are passed to the function
        params: custom parameters that will be passed to the function

    Returns
    --------
    fig : Matplotlib's figure object
    axes : list of Matplotlib's axes
    """

    scatter_kde_params.kde_show_shade = posterior_kde_params.data_kde_show_shade
    scatter_kde_params.kde_line_width = posterior_kde_params.data_kde_line_width

    fig, axes = scatter_and_kde(
                    values=values,
                    uncertainties=uncertainties,
                    title=title,
                    xlabel=xlabel,
                    ylabel=ylabel,
                    legend_labels=legend_labels,
                    scatter_kde_params=scatter_kde_params)

    kde_edge_colors = cycle(posterior_kde_params.kde_edgecolors)
    ax2 = axes[1]
    xlim = ax2.get_xlim()
    ylim = ax2.get_ylim()

    for fit in fits:
        kde_edge_color = next(kde_edge_colors)
        post = fit.get_drawset()
        x = np.linspace(xlim[0], xlim[1], posterior_kde_params.kde_values)

        for chain_id in range(fit.chains):
            nplots = int(posterior_kde_params.kde_plot_count / fit.chains)

            for i in range(nplots):
                row = post.iloc[chain_id * fit.draws + i]
                y = pdf(x=x, row=row)
                ax2.plot(x, y, color=kde_edge_color, zorder=1)

    ax2.set_xlim(xlim)

    # Ensure the new ylim upper value is not too large
    ylim_new = ax2.get_ylim()
    ylim_max = ylim[1] * min(2, ylim_new[1] / ylim[1])
    ax2.set_ylim(ylim[0], ylim_max)

    if plot_fn is not None:
        plot_fn[0](fig, axes, params=plot_fn[1])

    info_path.set_codefile()
    info_path.base_name = info_path.base_name or "posterior_scatter_kde"
    info_path.extension = info_path.extension or 'pdf'
    plot_path = get_info_path(info_path)
    fig.savefig(plot_path, dpi=info_path.dpi)

    return fig, axes



import pytest
from pytest import approx
import os
import shutil
import numpy as np
import pandas as pd
from tarpan.testutils.a03_cars.cars import get_fit

from tarpan.cmdstanpy.waic import (
    waic, compare_waic, save_compare_waic_csv, save_compare_waic_txt,
    waic_compared_to_df,
    WaicData, WaicModelCompared,
    compare_waic_tree_plot, save_compare_waic_tree_plot)

from tarpan.testutils.a04_height.height import (
    get_fit1_intercept, get_fit2_fungus_treatment, get_fit3_treatment)


def test_waic():
    fit = get_fit()
    result = waic(fit)

    assert result.waic == approx(421.5135196466395, rel=1e-15)

    assert len(result.waic_pointwise) == 50
    assert result.waic_pointwise[0] == approx(7.284060083431996, rel=1e-15)
    assert result.waic_pointwise[49] == approx(7.324510608904949, rel=1e-15)

    assert result.waic_std_err == approx(16.327468671341204, rel=1e-15)
    assert result.lppd == approx(-206.5875738029627, rel=1e-15)

    assert len(result.lppd_pointwise) == 50
    assert result.lppd_pointwise[0] == approx(-3.6203241203579615, rel=1e-15)
    assert result.lppd_pointwise[49] == approx(-3.641419673133626, rel=1e-15)

    assert result.penalty == approx(4.169186020357044, rel=1e-15)

    assert len(result.penalty_pointwise) == 50

    assert result.penalty_pointwise[0] == approx(0.021705921358036437,
                                                 rel=1e-15)

    assert result.penalty_pointwise[49] == approx(0.020835631318848448,
                                                  rel=1e-15)


def test_compare_waic():
    fit1_intercept = get_fit1_intercept()
    fit2_fungus_treatment = get_fit2_fungus_treatment()
    fit3_treatment = get_fit3_treatment()

    models = {
        "Itercept": fit1_intercept,
        "Fungus+treatment": fit2_fungus_treatment,
        "Treatment": fit3_treatment
    }

    result = compare_waic(models=models)

    assert [model.name for model in result] == ['Fungus+treatment',
                                                'Treatment',
                                                'Itercept']

    assert [round(model.waic_data.waic, 2) for model in result] == \
        [361.45, 402.71, 405.93]

    assert [round(model.waic_data.waic_std_err, 2) for model in result] == \
        [13.34, 10.78, 11.29]

    difference = [
        None if model.waic_difference_best is None
        else round(model.waic_difference_best, 2)
        for model in result
    ]

    assert difference == [None, 41.27, 44.48]

    std_err = [
        None if model.waic_difference_best_std_err is None
        else round(model.waic_difference_best_std_err, 2)
        for model in result
    ]

    assert std_err == [None, 9.82, 11.55]

    assert [round(model.waic_data.penalty, 1) for model in result] == \
        [3.4, 2.6, 1.6]

    actual_weight = [
        round(model.weight, 5)
        for model in result
    ]

    assert actual_weight == [0.99986, 2e-05, 0.00012]


def test_compare_waic__model_with_different_data_points():
    cars_fit = get_fit()
    plants_fit = get_fit1_intercept()

    models = {
        "Cars": cars_fit,
        "Plants": plants_fit
    }

    with pytest.raises(AttributeError,
                       match=r"different number of data points"):
        compare_waic(models=models)


def test_waic_compared_to_df():
    compared = []

    for i in range(1, 4):
        waic = WaicData(
            waic=i,
            waic_pointwise=[i] * 3,
            waic_std_err=i * 1.1,
            lppd=i * 1.2,
            lppd_pointwise=[i * 1.2] * 3,
            penalty=i * 0.3,
            penalty_pointwise=[i * 0.3] * 3,
        )

        compared_element = WaicModelCompared(
            name=f"Model {i}",
            waic_data=waic,
            waic_difference_best=i * 1.3,
            waic_difference_best_std_err=i * 1.4,
            weight=i*1.7
        )

        compared.append(compared_element)

    result = waic_compared_to_df(compared=compared)

    assert len(result) == 3

    row = result.loc["Model 1"]
    assert row["WAIC"] == 1
    assert row["SE"] == 1.1
    assert row["dWAIC"] == 1.3
    assert row["dSE"] == 1.4
    assert row["pWAIC"] == 0.3
    assert row["Weight"] == 1.7

    row = result.loc["Model 2"]
    assert row["WAIC"] == 2
    assert row["SE"] == 2.2
    assert row["dWAIC"] == 2.6
    assert row["dSE"] == 2.8
    assert row["pWAIC"] == 0.6
    assert row["Weight"] == 3.4


def test_save_compare_waic_csv():
    fit1_intercept = get_fit1_intercept()
    fit2_fungus_treatment = get_fit2_fungus_treatment()
    fit3_treatment = get_fit3_treatment()

    models = {
        "Itercept": fit1_intercept,
        "Fungus+treatment": fit2_fungus_treatment,
        "Treatment": fit3_treatment
    }

    outdir = "tarpan/cmdstanpy/model_info/waic_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_compare_waic_csv(models=models)

    assert os.path.isfile(os.path.join(outdir, "compare_waic.csv"))

    df = pd.read_csv(os.path.join(outdir, "compare_waic.csv"),
                     index_col="Name")

    assert len(df) == 3

    row = df.loc["Fungus+treatment"]
    assert row["WAIC"] == approx(361.44, rel=1e-3)
    assert row["SE"] == approx(13.33, rel=1e-3)
    assert np.isnan(row["dWAIC"])
    assert np.isnan(row["dSE"])
    assert row["pWAIC"] == approx(3.4388, rel=1e-3)
    assert row["Weight"] == approx(0.99985, rel=1e-3)

    row = df.loc["Itercept"]
    assert row["WAIC"] == approx(405.93, rel=1e-3)
    assert row["SE"] == approx(11.292, rel=1e-3)
    assert row["dWAIC"] == approx(44.48, rel=1e-3)
    assert row["dSE"] == approx(11.55, rel=1e-3)
    assert row["pWAIC"] == approx(1.5745, rel=1e-3)
    assert row["Weight"] == approx(0.00012332, rel=1e-3)


def test_save_compare_waic_txt():
    fit1_intercept = get_fit1_intercept()
    fit2_fungus_treatment = get_fit2_fungus_treatment()
    fit3_treatment = get_fit3_treatment()

    models = {
        "Itercept": fit1_intercept,
        "Fungus+treatment": fit2_fungus_treatment,
        "Treatment": fit3_treatment
    }

    outdir = "tarpan/cmdstanpy/model_info/waic_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_compare_waic_txt(models=models)

    assert os.path.isfile(os.path.join(outdir, "compare_waic.txt"))

    with open(os.path.join(outdir, "compare_waic.txt"), 'r') as file:
        data = file.read()
        assert "dWAIC" in data
        assert "Treatment" in data
        assert "402.71" in data


def test_compare_waic_tree_plot():
    fit1_intercept = get_fit1_intercept()
    fit2_fungus_treatment = get_fit2_fungus_treatment()
    fit3_treatment = get_fit3_treatment()

    models = {
        "Itercept": fit1_intercept,
        "Fungus+treatment": fit2_fungus_treatment,
        "Treatment": fit3_treatment
    }

    fig, ax = compare_waic_tree_plot(models=models)

    assert ax.get_xlabel() == "WAIC (deviance)"


def test_save_compare_waic_tree_plot():
    fit1_intercept = get_fit1_intercept()
    fit2_fungus_treatment = get_fit2_fungus_treatment()
    fit3_treatment = get_fit3_treatment()

    models = {
        "Itercept": fit1_intercept,
        "Fungus+treatment": fit2_fungus_treatment,
        "Treatment": fit3_treatment
    }

    outdir = "tarpan/cmdstanpy/model_info/waic_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_compare_waic_tree_plot(models=models)

    assert os.path.isfile(os.path.join(outdir, "compare_waic.pdf"))



from dataclasses import dataclass
from typing import List
import math
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from tabulate import tabulate
from scipy.special import logsumexp
from tarpan.cmdstanpy.waic import LPD_COLUMN_NAME_DEFAULT
from tarpan.cmdstanpy.psis_from_arviz import _psislw
from tarpan.shared.info_path import InfoPath, get_info_path
from tarpan.shared.tree_plot import tree_plot, TreePlotParams
from tarpan.shared.compare import model_weights


@dataclass
class ParetoKPlotParams:
    title: str = None  # Plot's title
    xlabel: str = "Point index"
    ylabel: str = "Pareto K"

    marker_color: str = "#0060ff44"
    marker_edgecolor: str = "#0060ff"

    # Colors for values above `max_pareto_k_value`
    marker_color_bad: str = "#ff002144"
    marker_edgecolor_bad: str = "#ff0021"

    marker: str = "o"

    grid_color: str = "#aaaaaa"
    grid_alpha: float = 0.2
    markersize: float = 80

    plot_width: float = 6
    plot_height: float = 6

    # Value at which a warning line is drawn
    max_pareto_k_value: float = 0.7

    # If True a text with index is shown next to a data points
    # with k values higher than `max_pareto_k_value`
    show_point_labels_for_high_k_points: bool = True

    point_label_size: float = 9

    # Values to be plotted on the x_axis. If none, plot point indixes 0, 1, 2...
    x_values: List[float] = None


# Results of PSIS calculations
@dataclass
class PsisData:
    """
    Compute PSIS (Pareto-smoothed importance sampling).
    It provides an estimate of models accuracy (out-of-the-sample deviance)
    and is used for comparing models. Smaller PSIS values are better.
    """
    psis: float

    """PSIS for each individual observation point."""
    psis_pointwise: List[float]

    """Standard error of PSIS (approximate)."""
    psis_std_err: float

    """
    LPPD (log-pointwise-predictive-density) value.
    LPPD is a measure of the accuracy of the model,
    larger is better. It describes how well the model corresponds to the
    observations.
    """
    lppd: float

    """LPPD for each individual observation point."""
    lppd_pointwise: List[float]

    """
    Penalty term, aka "effective number of parameters", which tells
    how much probabilities of observations vary across samples. The penalty
    The penlaty is added to PSIS number. The purpose of this is
    to combat overfitting, so models with too many parameters will have larger
    penlaties and larger PSIS.
    """
    penalty: float

    """Penatiles of all individual observations."""
    penalty_pointwise: List[float]

    """
    Pareto k values for each observation. Used as to diagnose the
    reliability of the estimation of the model accurace.
    If k > 0.7, then the observation is a possible outlier that influences
    the posterior distribution.
    """
    pareto_k: List[float]


@dataclass
class PsisModelCompared:
    """Model name"""
    name: str

    """Results of PSIS calculations for the model"""
    psis_data: PsisData

    """
    PSIS value difference between this model and the model with smallest psis.
    """
    psis_difference_best: float = None

    """
    Estimate of the standard error of the `psis_difference_best`.
    """
    psis_difference_best_std_err: float = None

    """
    Largest value of pareto k. Values above 0.7 indicate possible outliers.
    """
    largest_pareto_k: float = None

    """
    Appriximate measure of the relevance of the model, higher numbers
    correspond to models that are more compatible with the data.
    """
    weight: float = None


def compare_psis(models, lpd_column_name=LPD_COLUMN_NAME_DEFAULT) \
        -> List[PsisModelCompared]:
    """
    Compare models using PSIS (Pareto-smoothed importance sampling)
    to see which models are more compatible with the data.

    Parameters
    ----------

    models : dict
        key: str
            Model name.
        value: cmdstanpy.stanfit.CmdStanMCMC
            Contains the samples from cmdstanpy to compare.

    lpd_column_name : str
        Prefix of the columns in Stan's output that contain log
        probability density value for each observation. For example,
        if lpd_column_name='possum', when output is expected to have
        columns 'possum.1', 'possum.2', ..., 'possum.33' given 33 observations.

    Returns
    -------

    list of PsisModelCompared:
        List of PSIS comparisons. The list is sorted: models with
        lower PSIS falues (more compatible with data) come first.
    """

    psis_results = [
        PsisModelCompared(
            name=name,
            psis_data=psis(fit=fit, lpd_column_name=lpd_column_name)
        ) for name, fit in models.items()
    ]

    # Ensure all models have same number of observations
    # --------

    deviances_lengths = [
        len(result.psis_data.psis_pointwise) for result in psis_results
    ]

    if len(set(deviances_lengths)) > 1:
        raise AttributeError("Models have different number of data points")

    # Sort by PSIS, lower (better) first
    psis_results = sorted(psis_results, key=lambda x: x.psis_data.psis)
    best_model = psis_results[0]  # Model with smallest PSIS

    # Calculate model weights
    deviances = [result.psis_data.psis_pointwise for result in psis_results]
    weights = model_weights(deviances)

    # Calculate PSIS difference between models
    for i, model_result in enumerate(psis_results):
        n_points = len(model_result.psis_data.psis_pointwise)
        model_result.largest_pareto_k = max(model_result.psis_data.pareto_k)
        model_result.weight = weights[i]

        if i == 0:
            continue

        model_result.psis_difference_best = model_result.psis_data.psis - \
            best_model.psis_data.psis

        # Calculate standard error of the psis difference
        # ------------

        psis_difference_pointwise = \
            np.array(model_result.psis_data.psis_pointwise) - \
            np.array(best_model.psis_data.psis_pointwise)

        # Using the central limit theorem
        std_err = math.sqrt(n_points * psis_difference_pointwise.var())

        model_result.psis_difference_best_std_err = std_err

    return psis_results


def calculate_reff(fit):
    """
    Calculate number of effective samples divided by
    the number of actual samples.
    """

    df_summary = fit.summary()
    neff = float(df_summary[['N_Eff']].astype(int).mean())
    return neff / (fit.chains * fit.draws)


def psis(fit, lpd_column_name=LPD_COLUMN_NAME_DEFAULT) -> PsisData:
    """
    Compute PSIS (Pareto-smoothed importance sampling).
    It provides an estimate of models accuracy (out-of-the-sample deviance)
    and is used for comparing models. Smaller PSIS values are better,
    but they only make sense when compared.

    Parameters
    ----------

    fit : cmdstanpy.stanfit.CmdStanMCMC
        Contains the samples from cmdstanpy.

    lpd_column_name : str
        Prefix of the columns in Stan's output that contain log
        probability density value for each observation. For example,
        if lpd_column_name='possum', when output is expected to have
        columns 'possum.1', 'possum.2', ..., 'possum.33' given 33 observations.

    Returns
    -------

    PsisData:
        PSIS calculation result.
    """

    # Get log probability density of each observation.
    # The point_lpd is a 2D array with indexes:
    #   1. Sample
    #   2. Observation
    # ---------

    log_likelihood = fit.get_drawset(params=[lpd_column_name]).to_numpy()
    n_samples = log_likelihood.shape[0]  # Number of samples
    n_observations = log_likelihood.shape[1]  # Number of observations

    # Compute LPPD (log-pointwise-predictive-density) values for each
    # observation. LPPD is a measure of the accuracy of the model,
    # larger is better. It describes how well the model corresponds to the
    # observations.
    # ----------

    lppd_pointwise = [
        (logsumexp(log_likelihood[:, i]) - math.log(n_samples))
        for i in range(n_observations)
    ]

    lppd = sum(lppd_pointwise)  # Total LPPD

    # Calculed smoothed log_weights using PSIS
    # --------

    reff = calculate_reff(fit)
    log_likelihood_transposed = log_likelihood.transpose()
    log_weights, pareto_k = psislw(-log_likelihood_transposed, reff)
    log_weights += log_likelihood_transposed
    log_weights = log_weights.transpose()

    # Calculate PSIS for all observations
    # -------

    psis_pointwise = np.apply_along_axis(logsumexp, axis=0, arr=log_weights)

    # Compute penalty term, aka "effective number of parameters"
    # Variable `penalty_pointwise` is an array, each item corresponds
    # to one observation point.
    # -------

    penalty_pointwise = lppd_pointwise - psis_pointwise
    penalty = sum(penalty_pointwise)  # Total penalty

    # Convert PSIS from LPPD score to deviance
    psis_pointwise *= -2

    # Calculate total PSIS value
    psis = sum(psis_pointwise)

    # Approximate standard error of PSIS using the central limit theorem
    psis_std_err = math.sqrt(n_observations * psis_pointwise.var())

    result = PsisData(
        psis=psis,
        psis_pointwise=psis_pointwise,
        psis_std_err=psis_std_err,
        lppd=lppd,
        lppd_pointwise=lppd_pointwise,
        penalty=penalty,
        penalty_pointwise=penalty_pointwise,
        pareto_k=pareto_k
    )

    return result


def psislw(log_weights, reff=1.0):
    """
    Pareto smoothed importance sampling (PSIS).

    The function is taken from arviz library `psislw` function (stats.py)
    (https://github.com/arviz-devs/arviz), version 0.6.1,
    distributed under Apache License Version 2.0.

    Original code comes from Aki Vehtari, Tuomas Sivula:
      https://github.com/avehtari/PSIS

    Theory: https://arxiv.org/abs/1507.02646v5

    The original arviz function was modified by removing xarray functionality.


    Parameters
    ----------
    log_weights : array
        Array of size (n_observations, n_samples)
    reff : float
        relative MCMC efficiency, `ess / n`

    Returns
    -------
    lw_out : array
        Smoothed log weights
    kss : array
        Pareto tail indices (Pareto k values)
    """

    n_samples = log_weights.shape[-1]

    # precalculate constants
    cutoff_ind = -int(np.ceil(min(n_samples / 5.0, 3 * (n_samples / reff) ** 0.5))) - 1
    cutoffmin = np.log(np.finfo(float).tiny)
    k_min = 1.0 / 3

    result = np.apply_along_axis(_psislw, axis=1, arr=log_weights,
                                 cutoff_ind=cutoff_ind, cutoffmin=cutoffmin,
                                 k_min=k_min)

    log_weights = np.stack(result[:, 0], axis=0)
    pareto_k = result[:, 1]

    return log_weights, pareto_k


def psis_compared_to_df(compared: List[PsisModelCompared]):
    """
    Convert PSIS comparison to Pandas data frame.

    Parameters
    ----------
    compared: List[PsisModelCompared]
        Results of comparing PSIS between multiple models.

    Returns
    -------
    Pandas' DataFrame:
        PSIS comparison results.
    """

    column_names = ["PSIS", "SE", "dPSIS", "dSE", "pPSIS", "MaxK", "Weight"]
    model_names = [item.name for item in compared]
    df = pd.DataFrame(index=model_names, columns=column_names)

    for item in compared:
        psis = item.psis_data

        df.loc[item.name] = [
            psis.psis,
            psis.psis_std_err,
            item.psis_difference_best,
            item.psis_difference_best_std_err,
            psis.penalty,
            item.largest_pareto_k,
            item.weight
        ]

    return df


def save_compare_psis_csv(models,
                          lpd_column_name=LPD_COLUMN_NAME_DEFAULT,
                          info_path=InfoPath()):
    """
    Compare models using PSIS
    to see which models are more compatible with the data. The result
    is saved in a CSV file.

    Parameters
    ----------

    models : dict
        key: str
            Model name.
        value: cmdstanpy.stanfit.CmdStanMCMC
            Contains the samples from cmdstanpy to compare.

    lpd_column_name : str
        Prefix of the columns in Stan's output that contain log
        probability density value for each observation. For example,
        if lpd_column_name='possum', when output is expected to have
        columns 'possum.1', 'possum.2', ..., 'possum.33' given 33 observations.

    info_path : InfoPath
        Determines the location of the output file.
    """

    info_path.set_codefile()
    compared = compare_psis(models=models, lpd_column_name=lpd_column_name)
    save_compare_psis_csv_from_compared(compared=compared, info_path=info_path)


def save_compare_psis_csv_from_compared(compared, info_path=InfoPath()):
    """
    Compare models using PSIS
    to see which models are more compatible with the data. The result
    is saved in a CSV file.

    Parameters
    ----------

    compared : list of WaicModelCompared
        List of compared models.

    info_path : InfoPath
        Determines the location of the output file.
    """

    info_path.set_codefile()
    info_path = InfoPath(**info_path.__dict__)
    info_path.base_name = info_path.base_name or "compare_psis"
    info_path.extension = 'csv'
    df = psis_compared_to_df(compared)
    path = get_info_path(info_path)
    df.to_csv(path, index_label='Name')


def save_compare_psis_txt(models,
                          lpd_column_name=LPD_COLUMN_NAME_DEFAULT,
                          info_path=InfoPath()):
    """
    Compare models using PSIS
    to see which models are more compatible with the data. The result
    is saved in a text file.

    Parameters
    ----------

    models : dict
        key: str
            Model name.
        value: cmdstanpy.stanfit.CmdStanMCMC
            Contains the samples from cmdstanpy to compare.

    lpd_column_name : str
        Prefix of the columns in Stan's output that contain log
        probability density value for each observation. For example,
        if lpd_column_name='possum', when output is expected to have
        columns 'possum.1', 'possum.2', ..., 'possum.33' given 33 observations.

    info_path : InfoPath
        Determines the location of the output file.
    """

    info_path.set_codefile()
    compared = compare_psis(models=models, lpd_column_name=lpd_column_name)
    save_compare_psis_txt_from_compared(compared=compared, info_path=info_path)


def save_compare_psis_txt_from_compared(compared, info_path=InfoPath()):
    """
    Compare models using PSIS
    to see which models are more compatible with the data. The result
    is saved in a text file.

    Parameters
    ----------

    compared : list of WaicModelCompared
        List of compared models.

    info_path : InfoPath
        Determines the location of the output file.
    """

    info_path.set_codefile()
    info_path = InfoPath(**info_path.__dict__)
    info_path.base_name = info_path.base_name or "compare_psis"
    info_path.extension = 'txt'
    df = psis_compared_to_df(compared)
    table = tabulate(df, headers=list(df), floatfmt=".2f", tablefmt="pipe")
    path = get_info_path(info_path)

    with open(path, "w") as text_file:
        print(table, file=text_file)


def compare_psis_tree_plot(models, lpd_column_name=LPD_COLUMN_NAME_DEFAULT,
                           tree_plot_params: TreePlotParams = TreePlotParams()):
    """
    Make a plot that compares models using PSIS.

    Parameters
    ----------

    models : dict
        key: str
            Model name.
        value: cmdstanpy.stanfit.CmdStanMCMC
            Contains the samples from cmdstanpy to compare.

    lpd_column_name : str
        Prefix of the columns in Stan's output that contain log
        probability density value for each observation. For example,
        if lpd_column_name='possum', when output is expected to have
        columns 'possum.1', 'possum.2', ..., 'possum.33' given 33 observations.

    Returns
    -------
    (fig, ax):
        fig: Matplotlib's figure
        ax: Matplotlib's axis
    """

    compared = compare_psis(models=models, lpd_column_name=lpd_column_name)

    return compare_psis_tree_plot_from_compared(
        compared=compared, tree_plot_params=tree_plot_params)


def compare_psis_tree_plot_from_compared(
        compared, tree_plot_params: TreePlotParams = TreePlotParams()):
    """
    Make a plot that compares models using PSIS.

    Parameters
    ----------

    compared : list of WaicModelCompared
        List of compared models.

    Returns
    -------
    (fig, ax):
        fig: Matplotlib's figure
        ax: Matplotlib's axis
    """

    plot_groups = []
    tree_plot_params = TreePlotParams(**tree_plot_params.__dict__)

    if tree_plot_params.labels is None:
        tree_plot_params.labels = ["dPSIS", "PSIS"]

    if tree_plot_params.xlabel is None:
        tree_plot_params.xlabel = "PSIS (deviance)"

    if tree_plot_params.title is None:
        tree_plot_params.title = "Model comparison (smaller is better)"

    for model in reversed(compared):
        values = []
        psis = model.psis_data
        group = dict(name=model.name, values=values)
        plot_groups.append(group)

        # PSIS difference
        # --------

        value = dict(value=psis.psis, error_bars=[])

        if model.psis_difference_best_std_err is not None:
            error_bars = [
                psis.psis - model.psis_difference_best_std_err,
                psis.psis + model.psis_difference_best_std_err
            ]

            value = dict(value=psis.psis, error_bars=[error_bars])

        values.append(value)

        # PSIS value
        # --------

        error_bars = [
            psis.psis - psis.psis_std_err,
            psis.psis + psis.psis_std_err
        ]

        value = dict(value=psis.psis, error_bars=[error_bars])

        values.append(value)

    tree_plot_params.draw_zero_line_if_in_range = False
    fig, ax = tree_plot(groups=plot_groups, params=tree_plot_params)

    # Draw a vertical line through the best model
    # ----------

    model = compared[0]

    ax.axvline(x=model.psis_data.psis, linestyle='dashed',
               color=tree_plot_params.marker_edge_colors[0])

    return fig, ax


def save_compare_psis_tree_plot(
        models, lpd_column_name=LPD_COLUMN_NAME_DEFAULT,
        tree_plot_params: TreePlotParams = TreePlotParams(),
        info_path=InfoPath()):
    """
    Make a plot that compares models using PSIS
    and save it to a file.

    Parameters
    ----------

    models : dict
        key: str
            Model name.
        value: cmdstanpy.stanfit.CmdStanMCMC
            Contains the samples from cmdstanpy to compare.

    lpd_column_name : str
        Prefix of the columns in Stan's output that contain log
        probability density value for each observation. For example,
        if lpd_column_name='possum', when output is expected to have
        columns 'possum.1', 'possum.2', ..., 'possum.33' given 33 observations.


    info_path : InfoPath
        Determines the location of the output file.
    """

    info_path.set_codefile()
    compared = compare_psis(models=models, lpd_column_name=lpd_column_name)

    save_compare_psis_tree_plot_from_compared(
        compared=compared,
        tree_plot_params=tree_plot_params,
        info_path=info_path)


def save_compare_psis_tree_plot_from_compared(
        compared,
        tree_plot_params: TreePlotParams = TreePlotParams(),
        info_path=InfoPath()):
    """
    Make a plot that compares models using PSIS
    and save it to a file.

    Parameters
    ----------

    compared : list of WaicModelCompared
        List of compared models.

    info_path : InfoPath
        Determines the location of the output file.
    """

    info_path.set_codefile()
    info_path = InfoPath(**info_path.__dict__)

    fig, ax = compare_psis_tree_plot_from_compared(
        compared=compared,
        tree_plot_params=tree_plot_params)

    info_path.base_name = info_path.base_name or 'compare_psis'
    info_path.extension = info_path.extension or 'pdf'
    the_path = get_info_path(info_path)
    fig.savefig(the_path, dpi=info_path.dpi)
    plt.close(fig)


def psis_pareto_k_plot(fit, name,
                       lpd_column_name=LPD_COLUMN_NAME_DEFAULT,
                       pareto_k_plot_params: ParetoKPlotParams = ParetoKPlotParams()):
    """
    Make a plot that shows values of Pareto K index generated by PSIS
    method. This is used to see if there are values of K higher than 0.7,
    which means that there are possible outliers and PSIS calculations
    may not be reliable.

    Parameters
    ----------

    fit : cmdstanpy.stanfit.CmdStanMCMC
        Contains the samples from cmdstanpy.

    name : str
        Model name.

    lpd_column_name : str
        Prefix of the columns in Stan's output that contain log
        probability density value for each observation. For example,
        if lpd_column_name='possum', when output is expected to have
        columns 'possum.1', 'possum.2', ..., 'possum.33' given 33 observations.

    Returns
    -------
    (fig, ax):
        fig: Matplotlib's figure
        ax: Matplotlib's axis
    """

    psis_data = psis(fit=fit, lpd_column_name=lpd_column_name)

    return psis_pareto_k_plot_from_psis_data(
            psis_data=psis_data, name=name,
            pareto_k_plot_params=pareto_k_plot_params)


def psis_pareto_k_plot_from_psis_data(
        psis_data: PsisData, name,
        pareto_k_plot_params: ParetoKPlotParams = ParetoKPlotParams()):
    """
    Make a plot that shows values of Pareto K index generated by PSIS
    method. This is used to see if there are values of K higher than 0.7,
    which means that there are possible outliers and PSIS calculations
    may not be reliable.

    Parameters
    ----------

    psis_data : PsisData
        Calculated PSIS values for the model.

    name : str
        Model name.

    Returns
    -------
    (fig, ax):
        fig: Matplotlib's figure
        ax: Matplotlib's axis
    """

    k_values = psis_data.pareto_k

    fig, ax = plt.subplots(1, 1,
                           figsize=(pareto_k_plot_params.plot_width,
                                    pareto_k_plot_params.plot_height))

    x_values = pareto_k_plot_params.x_values

    if x_values is None:
        x_values = list(range(len(k_values)))

    x_values = np.array(x_values)
    k_values = np.array(k_values)

    # Plot 'good' values (below max_pareto_k_value)
    # ------

    x = x_values[k_values <= pareto_k_plot_params.max_pareto_k_value]
    y = k_values[k_values <= pareto_k_plot_params.max_pareto_k_value]

    ax.scatter(x, y,
               marker=pareto_k_plot_params.marker,
               color=pareto_k_plot_params.marker_color,
               edgecolor=pareto_k_plot_params.marker_edgecolor,
               s=pareto_k_plot_params.markersize)

    # Plot 'bad' values (above max_pareto_k_value)
    # ------

    x = x_values[k_values > pareto_k_plot_params.max_pareto_k_value]
    y = k_values[k_values > pareto_k_plot_params.max_pareto_k_value]

    ax.scatter(x, y,
               marker=pareto_k_plot_params.marker,
               color=pareto_k_plot_params.marker_color_bad,
               edgecolor=pareto_k_plot_params.marker_edgecolor_bad,
               s=pareto_k_plot_params.markersize)

    ax.axhline(y=pareto_k_plot_params.max_pareto_k_value,
               linestyle='dashed',
               color=pareto_k_plot_params.marker_edgecolor_bad)

    # Determine marker size in data coordinates
    # This will be used to place text below the markers
    # ------

    # There are 72 points per inch.
    points_per_pixel = 72. / ax.figure.dpi
    trans = ax.transData.inverted()
    marker_size_in_points = math.sqrt(pareto_k_plot_params.markersize)
    marker_size_in_pixels = marker_size_in_points / points_per_pixel

    marker_size_in_data_coord = trans.transform(
        (marker_size_in_pixels, marker_size_in_pixels)) -\
        trans.transform((0, 0))

    if pareto_k_plot_params.show_point_labels_for_high_k_points:
        for i, k in enumerate(k_values):
            if k < pareto_k_plot_params.max_pareto_k_value:
                continue

            # Adjust y position of text to place it below the marker
            y = k - marker_size_in_data_coord[1] * 1.1

            ax.text(x_values[i], y, i,
                    horizontalalignment='center',
                    verticalalignment='center',
                    size=pareto_k_plot_params.point_label_size,
                    color=pareto_k_plot_params.marker_edgecolor_bad)

    if pareto_k_plot_params.xlabel is not None:
        ax.set_xlabel(pareto_k_plot_params.xlabel)

    if pareto_k_plot_params.ylabel is not None:
        ax.set_ylabel(pareto_k_plot_params.ylabel)

    title = pareto_k_plot_params.title

    if title is None:
        title = f"'{name}' Pareto K values"

    ax.set_title(title)

    ax.grid(color=pareto_k_plot_params.grid_color, linewidth=1,
            alpha=pareto_k_plot_params.grid_alpha)

    fig.tight_layout()

    return fig, ax


def save_psis_pareto_k_plot(
        fit, name,
        lpd_column_name=LPD_COLUMN_NAME_DEFAULT,
        pareto_k_plot_params: ParetoKPlotParams = ParetoKPlotParams(),
        info_path=InfoPath()):
    """
    Make a plot that shows values of Pareto K index generated by PSIS
    method. This is used to see if there are values of K higher than 0.7,
    which means that there are possible outliers and PSIS calculations
    may not be reliable.

    Parameters
    ----------

    fit : cmdstanpy.stanfit.CmdStanMCMC
        Contains the samples from cmdstanpy.

    name : str
        Model name.

    info_path : InfoPath
        Determines the location of the output file.
    """

    info_path.set_codefile()
    psis_data = psis(fit=fit, lpd_column_name=lpd_column_name)

    save_psis_pareto_k_plot_from_psis_data(
        psis_data=psis_data, name=name,
        pareto_k_plot_params=pareto_k_plot_params,
        info_path=info_path)


def save_psis_pareto_k_plot_from_psis_data(
        psis_data: PsisData, name,
        pareto_k_plot_params: ParetoKPlotParams = ParetoKPlotParams(),
        info_path=InfoPath()):
    """
    Make a plot that shows values of Pareto K index generated by PSIS
    method. This is used to see if there are values of K higher than 0.7,
    which means that there are possible outliers and PSIS calculations
    may not be reliable.

    Parameters
    ----------

    psis_data : PsisData
        Calculated PSIS values for the model.

    name : str
        Model name.

    info_path : InfoPath
        Determines the location of the output file.
    """

    info_path.set_codefile()
    info_path = InfoPath(**info_path.__dict__)

    fig, ax = psis_pareto_k_plot_from_psis_data(
        psis_data=psis_data, name=name,
        pareto_k_plot_params=pareto_k_plot_params)

    repalce_characters = [" ", "\\", "/", "?", "+", "*"]
    model_name_sanitised = name.lower()

    for character in repalce_characters:
        model_name_sanitised = model_name_sanitised.replace(character, "_")

    base_name = f'pareto_k_{model_name_sanitised}'
    info_path.base_name = info_path.base_name or base_name
    info_path.extension = info_path.extension or 'pdf'
    the_path = get_info_path(info_path)
    fig.savefig(the_path, dpi=info_path.dpi)
    plt.close(fig)


def save_psis_pareto_k_plot_from_compared(
        compared,
        pareto_k_plot_params: ParetoKPlotParams = ParetoKPlotParams(),
        info_path=InfoPath()):
    """
    Make multiple plots that show values of Pareto K index generated by PSIS
    method for each compared model.

    Parameters
    ----------

    compared : list of WaicModelCompared
        List of compared models.

    info_path : InfoPath
        Determines the location of the output file.
    """

    info_path.set_codefile()

    for model in compared:
        psis_data = model.psis_data

        save_psis_pareto_k_plot_from_psis_data(
            psis_data=psis_data, name=model.name,
            pareto_k_plot_params=pareto_k_plot_params,
            info_path=info_path)



import numpy as np
import scipy.stats as st


def model_weights(deviances, b_samples=1000):
    """
    Calculate weights of the model that can be used to compare them.

    Pseudo-Bayesian Model averaging using Akaike-type
    weighting. The weights are stabilized using the Bayesian bootstrap.

    The code is taken from `compare` function of stats.py file
    of arviz library (https://github.com/arviz-devs/arviz), version 0.6.1,
    distributed under Apache License Version 2.0.

    Parameters
    ----------
    deviances: 2D array
        An array containing WAIC or PSIS values for each model and observations
        Rows are models and columns are observations.

    b_samples: int
        Number of samples taken by the Bayesian bootstrap estimation.

    Returns
    -------
    list:
        Weight of each of the model. Higher value means the model
        is more compatible with the data.
    """

    ic_i_val = np.array(deviances).transpose()
    rows = ic_i_val.shape[0]
    cols = ic_i_val.shape[1]
    ic_i_val = ic_i_val * rows

    b_weighting = st.dirichlet.rvs(alpha=[1] * rows, size=b_samples,
                                   random_state=1)

    weights = np.zeros((b_samples, cols))
    z_bs = np.zeros_like(weights)

    for i in range(b_samples):
        z_b = np.dot(b_weighting[i], ic_i_val)
        u_weights = np.exp((z_b - np.min(z_b)) / -2)
        z_bs[i] = z_b
        weights[i] = u_weights / np.sum(u_weights)

    return weights.mean(axis=0).tolist()



"""Makes statistical summary of parameter destibutions: means, std etc."""

import numpy as np
from tarpan.shared.info_path import InfoPath

from tarpan.shared.summary import (
    SummaryParams, sample_summary, save_summary_to_disk)

from tarpan.shared.param_names import filter_param_names


def save_summary(fit, param_names=None, info_path=InfoPath(),
                 summary_params=SummaryParams()):
    """
    Saves statistical summary of the samples using mean, std, mode, hpdi.

    Parameters
    ----------

    fit : cmdstanpy.stanfit.CmdStanMCMC

        Contains the samples from cmdstanpy.

    param_names : list of str

        Names of parameters to be included in the summary. Include all if None.

    info_path : InfoPath

        Path information for creating summaries.

    Returns
    -------
    dict:
        df:
            Panda's data frame containing the summary
        table: str
            Summary table in text format.
        samples: Panda's data frame
            Combined samples from all chains
        path_txt: str
            Path to the text summary
        path_csv: str
            Path to summary in CSV format
    """

    info_path.set_codefile()
    info_path = InfoPath(**info_path.__dict__)

    df_summary, summary, samples = make_summary(
        fit, param_names=param_names, summary_params=summary_params)

    output = save_summary_to_disk(df_summary, summary, info_path=info_path)

    return {
        "df": df_summary,
        "table": summary,
        "samples": samples,
        "path_txt": output["path_txt"],
        "path_csv": output["path_csv"]
    }


def make_summary(fit, param_names, summary_params=SummaryParams()):
    """
    Returns statistical summary table for parameters:
    mean, std, mode, hpdi.

    Parameters
    ----------

    fit : cmdstanpy.stanfit.CmdStanMCMC

        Contains the samples from cmdstanpy.

    param_names : list of str

        Names of parameters to be included in the summar. Include all if None.
    """

    param_names = filter_param_names(fit.column_names, param_names)
    samples = fit.get_drawset(params=param_names)

    # Get R_hat values from the summary
    # --------

    df_summary = fit.summary()

    df_summary.rename(
        index=(lambda name: name.replace('[', '.').replace(']', '')),
        inplace=True)

    df_summary = df_summary[['N_Eff', 'R_hat']]
    df_summary['N_Eff'] = np.round(df_summary['N_Eff'])
    df_summary['N_Eff'] = df_summary['N_Eff'].astype(int)

    # Get the summary
    df_summary, table = sample_summary(df=samples, extra_values=df_summary,
                                       params=summary_params)

    return df_summary, table, samples



import math
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from dataclasses import dataclass
from typing import List
from tabulate import tabulate
from scipy.special import logsumexp
from tarpan.shared.info_path import InfoPath, get_info_path
from tarpan.shared.tree_plot import tree_plot, TreePlotParams
from tarpan.shared.compare import model_weights


"""
Prefix of the columns in Stan's output that contain log
probability density value for each observation. For example,
if lpd_column_name='possum', when output is expected to have
columns 'possum.1', 'possum.2', ..., 'possum.33' given 33 observations.
"""
LPD_COLUMN_NAME_DEFAULT = "lpd_pointwise"


# Results of WAIC calculations
@dataclass
class WaicData:
    """
    WAIC value (Widely Aplicable Information Criterion).
    It provides an estimate of models accuracy (out-of-the-sample deviance)
    and is used for comparing models. Smaller WAIC values are better.
    """
    waic: float

    """WAIC for each individual observation point."""
    waic_pointwise: List[float]

    """Standard error of WAIC (approximate)."""
    waic_std_err: float

    """
    LPPD (log-pointwise-predictive-density) value.
    LPPD is a measure of the accuracy of the model,
    larger is better. It describes how well the model corresponds to the
    observations.
    """
    lppd: float

    """LPPD for each individual observation point."""
    lppd_pointwise: List[float]

    """
    Penalty term, aka "effective number of parameters", which tells
    how much probabilities of observations vary across samples. The penalty
    The penalty is added to WAIC number. The purpose of this is
    to combat overfitting, so models with too many parameters will have larger
    penlties and larger WAIC.
    """
    penalty: float

    """Penatiles of all individual observations."""
    penalty_pointwise: List[float]


@dataclass
class WaicModelCompared:
    """Model name"""
    name: str

    """Results of WAIC calculations for the model"""
    waic_data: WaicData

    """
    waic value difference between this model and the model with smallest waic.
    """
    waic_difference_best: float = None

    """
    Estimate of the standard error of the `waic_difference_best`.
    """
    waic_difference_best_std_err: float = None

    """
    Approximate measure of the relevance of the model, higher numbers
    correspond to models that are more compatible with the data.
    """
    weight: float = None


def waic(fit, lpd_column_name=LPD_COLUMN_NAME_DEFAULT) -> WaicData:
    """
    Compute WAIC (Widely Aplicable Information Criterion).
    It provides an estimate of models accuracy (out-of-the-sample deviance)
    and is used for comparing models. Smaller WAIC values are better,
    but they only make sense when compared.

    Parameters
    ----------

    fit : cmdstanpy.stanfit.CmdStanMCMC
        Contains the samples from cmdstanpy.

    lpd_column_name : str
        Prefix of the columns in Stan's output that contain log
        probability density value for each observation. For example,
        if lpd_column_name='possum', when output is expected to have
        columns 'possum.1', 'possum.2', ..., 'possum.33' given 33 observations.

    Returns
    -------

    WaicData:
        Waic calculation result.
    """

    # Get log probability density of each observation.
    # The point_lpd is a 2D array with indexes:
    #   1. Sample
    #   2. Observation
    # ---------

    lpd_pointwise = fit.get_drawset(params=[lpd_column_name])
    n_samples = lpd_pointwise.shape[0]  # Number of samples
    n_observations = lpd_pointwise.shape[1]  # Number of observations

    # Compute LPPD (log-pointwise-predictive-density) values for each
    # observation. LPPD is a measure of the accuracy of the model,
    # larger is better. It describes how well the model corresponds to the
    # observations.
    # ----------

    lppd_pointwise = [
        (logsumexp(lpd_pointwise.iloc[:, i]) - math.log(n_samples))
        for i in range(n_observations)
    ]

    # Compute penalty term, aka "effective number of parameters"
    # Variable `penalty_pointwise` is an array, each item corresponds
    # to one observation point. Each item contains variance of log probability
    # density values from all samples for a single observation.

    penalty_pointwise = [
        lpd_pointwise.iloc[:, i].var()
        for i in range(n_observations)
    ]

    # Compute WAIC
    # -------

    penalty = sum(penalty_pointwise)
    lppd = sum(lppd_pointwise)
    waic = -2*(lppd - penalty)

    # Approximate standard error of WAIC using the central limit theorem
    # -------

    waic_pointwise = -2*(np.array(lppd_pointwise) -
                         np.array(penalty_pointwise))

    waic_std_err = math.sqrt(n_observations * waic_pointwise.var())

    result = WaicData(
        waic=waic,
        waic_pointwise=waic_pointwise,
        waic_std_err=waic_std_err,
        lppd=lppd,
        lppd_pointwise=lppd_pointwise,
        penalty=penalty,
        penalty_pointwise=penalty_pointwise,
    )

    return result


def compare_waic(models, lpd_column_name=LPD_COLUMN_NAME_DEFAULT) \
        -> List[WaicModelCompared]:
    """
    Compare models using WAIC (Widely Aplicable Information Criterion)
    to see which models are more compatible with the data.

    Parameters
    ----------

    models : dict
        key: str
            Model name
        value: cmdstanpy.stanfit.CmdStanMCMC
            Contains the samples from cmdstanpy to compare

    lpd_column_name : str
        Prefix of the columns in Stan's output that contain log
        probability density value for each observation. For example,
        if lpd_column_name='possum', when output is expected to have
        columns 'possum.1', 'possum.2', ..., 'possum.33' given 33 observations.

    Returns
    -------

    list of WaicModelCompared:
        List of WAIC comparisons. The list is sorted: models with
        lower WAIC falues (more compatible with data) come first.
    """

    waic_results = [
        WaicModelCompared(
            name=name,
            waic_data=waic(fit=fit, lpd_column_name=lpd_column_name)
        ) for name, fit in models.items()
    ]

    # Ensure all models have same number of observations
    # --------

    deviances_lengths = [
        len(result.waic_data.waic_pointwise) for result in waic_results
    ]

    if len(set(deviances_lengths)) > 1:
        raise AttributeError("Models have different number of data points")

    # Sort by WAIC, lower (better) first
    waic_results = sorted(waic_results, key=lambda x: x.waic_data.waic)

    best_model = waic_results[0]  # Model with smallest WAIC

    # Calculate model weights
    deviances = [result.waic_data.waic_pointwise for result in waic_results]
    weights = model_weights(deviances)

    # Calculate WAIC difference between models
    for i, model_result in enumerate(waic_results):
        n_points = len(model_result.waic_data.waic_pointwise)
        model_result.weight = weights[i]

        if i == 0:
            continue

        model_result.waic_difference_best = model_result.waic_data.waic - \
            best_model.waic_data.waic

        # Calculate standard error of the waic difference
        # ------------

        waic_difference_pointwise = \
            np.array(model_result.waic_data.waic_pointwise) - \
            np.array(best_model.waic_data.waic_pointwise)

        # Using the central limit theorem
        std_err = math.sqrt(n_points * waic_difference_pointwise.var())

        model_result.waic_difference_best_std_err = std_err

    return waic_results


def waic_compared_to_df(compared: List[WaicModelCompared]):
    """
    Convert waic comparison to Pandas data frame.

    Parameters
    ----------
    compared: List[WaicModelCompared]
        Results of comparing WAIC between multiple models.

    Returns
    -------
    Pandas' DataFrame:
        WAIC comparison results.
    """

    column_names = ["WAIC", "SE", "dWAIC", "dSE", "pWAIC", "Weight"]
    model_names = [item.name for item in compared]
    df = pd.DataFrame(index=model_names, columns=column_names)

    for item in compared:
        waic = item.waic_data

        df.loc[item.name] = [
            waic.waic,
            waic.waic_std_err,
            item.waic_difference_best,
            item.waic_difference_best_std_err,
            waic.penalty,
            item.weight
        ]

    return df


def save_compare_waic_csv(models,
                          lpd_column_name=LPD_COLUMN_NAME_DEFAULT,
                          info_path=InfoPath()):
    """
    Compare models using WAIC (Widely Aplicable Information Criterion)
    to see which models are more compatible with the data. The result
    is saved in a CSV file.

    Parameters
    ----------

    models : dict
        key: str
            Model name.
        value: cmdstanpy.stanfit.CmdStanMCMC
            Contains the samples from cmdstanpy to compare.

    lpd_column_name : str
        Prefix of the columns in Stan's output that contain log
        probability density value for each observation. For example,
        if lpd_column_name='possum', when output is expected to have
        columns 'possum.1', 'possum.2', ..., 'possum.33' given 33 observations.

    info_path : InfoPath
        Determines the location of the output file.
    """

    info_path.set_codefile()
    compared = compare_waic(models=models, lpd_column_name=lpd_column_name)
    save_compare_waic_csv_from_compared(compared=compared, info_path=info_path)


def save_compare_waic_csv_from_compared(compared, info_path=InfoPath()):
    """
    Compare models using WAIC (Widely Aplicable Information Criterion)
    to see which models are more compatible with the data. The result
    is saved in a CSV file.

    Parameters
    ----------

    compared : list of WaicModelCompared
        List of compared models.
    info_path : InfoPath
        Determines the location of the output file.
    """

    info_path.set_codefile()
    info_path = InfoPath(**info_path.__dict__)
    info_path.base_name = info_path.base_name or "compare_waic"
    info_path.extension = 'csv'
    df = waic_compared_to_df(compared)
    path = get_info_path(info_path)
    df.to_csv(path, index_label='Name')


def save_compare_waic_txt(models,
                          lpd_column_name=LPD_COLUMN_NAME_DEFAULT,
                          info_path=InfoPath()):
    """
    Compare models using WAIC (Widely Aplicable Information Criterion)
    to see which models are more compatible with the data. The result
    is saved in a text file.

    Parameters
    ----------

    models : dict
        key: str
            Model name.
        value: cmdstanpy.stanfit.CmdStanMCMC
            Contains the samples from cmdstanpy to compare.

    lpd_column_name : str
        Prefix of the columns in Stan's output that contain log
        probability density value for each observation. For example,
        if lpd_column_name='possum', when output is expected to have
        columns 'possum.1', 'possum.2', ..., 'possum.33' given 33 observations.

    info_path : InfoPath
        Determines the location of the output file.
    """

    info_path.set_codefile()
    compared = compare_waic(models=models, lpd_column_name=lpd_column_name)
    save_compare_waic_txt_from_compared(compared=compared, info_path=info_path)


def save_compare_waic_txt_from_compared(compared, info_path=InfoPath()):
    """
    Compare models using WAIC (Widely Aplicable Information Criterion)
    to see which models are more compatible with the data. The result
    is saved in a text file.

    Parameters
    ----------

    compared : list of WaicModelCompared
        List of compared models.

    info_path : InfoPath
        Determines the location of the output file.
    """

    info_path.set_codefile()
    info_path = InfoPath(**info_path.__dict__)
    info_path.base_name = info_path.base_name or "compare_waic"
    info_path.extension = 'txt'
    df = waic_compared_to_df(compared)
    table = tabulate(df, headers=list(df), floatfmt=".2f", tablefmt="pipe")
    path = get_info_path(info_path)

    with open(path, "w") as text_file:
        print(table, file=text_file)


def compare_waic_tree_plot(models, lpd_column_name=LPD_COLUMN_NAME_DEFAULT,
                           tree_plot_params: TreePlotParams = TreePlotParams()):
    """
    Make a plot that compares models using WAIC
    (Widely Aplicable Information Criterion).

    Parameters
    ----------

    models : dict
        key: str
            Model name.
        value: cmdstanpy.stanfit.CmdStanMCMC
            Contains the samples from cmdstanpy to compare.

    lpd_column_name : str
        Prefix of the columns in Stan's output that contain log
        probability density value for each observation. For example,
        if lpd_column_name='possum', when output is expected to have
        columns 'possum.1', 'possum.2', ..., 'possum.33' given 33 observations.

    Returns
    -------
    (fig, ax):
        fig: Matplotlib's figure
        ax: Matplotlib's axis
    """

    compared = compare_waic(models=models, lpd_column_name=lpd_column_name)

    return compare_waic_tree_plot_from_compared(
            compared=compared, tree_plot_params=tree_plot_params)


def compare_waic_tree_plot_from_compared(
        compared, tree_plot_params: TreePlotParams = TreePlotParams()):
    """
    Make a plot that compares models using WAIC
    (Widely Aplicable Information Criterion).

    Parameters
    ----------

    compared : list of WaicModelCompared
        List of compared models.

    Returns
    -------
    (fig, ax):
        fig: Matplotlib's figure
        ax: Matplotlib's axis
    """

    plot_groups = []
    tree_plot_params = TreePlotParams(**tree_plot_params.__dict__)

    if tree_plot_params.labels is None:
        tree_plot_params.labels = ["dWAIC", "WAIC"]

    if tree_plot_params.xlabel is None:
        tree_plot_params.xlabel = "WAIC (deviance)"

    if tree_plot_params.title is None:
        tree_plot_params.title = "Model comparison (smaller is better)"

    for model in reversed(compared):
        values = []
        waic = model.waic_data
        group = dict(name=model.name, values=values)
        plot_groups.append(group)

        # WAIC difference
        # --------

        value = dict(value=waic.waic, error_bars=[])

        if model.waic_difference_best_std_err is not None:
            error_bars = [
                waic.waic - model.waic_difference_best_std_err,
                waic.waic + model.waic_difference_best_std_err
            ]

            value = dict(value=waic.waic, error_bars=[error_bars])

        values.append(value)

        # WAIC value
        # --------

        error_bars = [
            waic.waic - waic.waic_std_err,
            waic.waic + waic.waic_std_err
        ]

        value = dict(value=waic.waic, error_bars=[error_bars])

        values.append(value)

    tree_plot_params.draw_zero_line_if_in_range = False
    fig, ax = tree_plot(groups=plot_groups, params=tree_plot_params)

    # Draw a vertical line through the best model
    # ----------

    model = compared[0]

    ax.axvline(x=model.waic_data.waic, linestyle='dashed',
               color=tree_plot_params.marker_edge_colors[0])

    return fig, ax


def save_compare_waic_tree_plot(
        models, lpd_column_name=LPD_COLUMN_NAME_DEFAULT,
        tree_plot_params: TreePlotParams = TreePlotParams(),
        info_path=InfoPath()):
    """
    Make a plot that compares models using WAIC
    (Widely Aplicable Information Criterion) and save it to a file.

    Parameters
    ----------

    models : list of dict
        List of model samples from cmdstanpy to compare.

        The dictionary has keys:
            name: str
                Model name
            fit: cmdstanpy.stanfit.CmdStanMCMC
                Contains the samples from cmdstanpy.

    lpd_column_name : str
        Prefix of the columns in Stan's output that contain log
        probability density value for each observation. For example,
        if lpd_column_name='possum', when output is expected to have
        columns 'possum.1', 'possum.2', ..., 'possum.33' given 33 observations.


    info_path : InfoPath
        Determines the location of the output file.
    """

    info_path.set_codefile()
    compared = compare_waic(models=models, lpd_column_name=lpd_column_name)

    save_compare_waic_tree_plot_from_compared(
        compared=compared,
        tree_plot_params=tree_plot_params,
        info_path=info_path)


def save_compare_waic_tree_plot_from_compared(
        compared,
        tree_plot_params: TreePlotParams = TreePlotParams(),
        info_path=InfoPath()):
    """
    Make a plot that compares models using WAIC
    (Widely Aplicable Information Criterion) and save it to a file.

    Parameters
    ----------

    compared : list of WaicModelCompared
        List of compared models.

    lpd_column_name : str
        Prefix of the columns in Stan's output that contain log
        probability density value for each observation. For example,
        if lpd_column_name='possum', when output is expected to have
        columns 'possum.1', 'possum.2', ..., 'possum.33' given 33 observations.


    info_path : InfoPath
        Determines the location of the output file.
    """

    info_path.set_codefile()
    info_path = InfoPath(**info_path.__dict__)

    fig, ax = compare_waic_tree_plot_from_compared(
        compared=compared,
        tree_plot_params=tree_plot_params)

    info_path.base_name = info_path.base_name or 'compare_waic'
    info_path.extension = info_path.extension or 'pdf'
    the_path = get_info_path(info_path)
    fig.savefig(the_path, dpi=info_path.dpi)
    plt.close(fig)



import shutil
import os
import pandas as pd
from tarpan.shared.analyse import save_analysis


def test_save_analysis():
    data = {
        "x": [1, 2, 3, 4, 5, 6, 2, 3, 4, 4, 1, 3, 4, 5, 6, 2],
        "y": [-1, -2, -3, -4, -5, -6, -2, -2, -1, -1, -1, -3, -3, -2, -1, -4]
    }

    df = pd.DataFrame(data)

    outdir = "tarpan/shared/model_info/analyse_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_analysis(df)

    # Check summary files
    # ---------------

    assert os.path.isfile(os.path.join(outdir, "summary.txt"))
    assert os.path.isfile(os.path.join(outdir, "summary.csv"))

    # Tree plot
    assert os.path.isfile(os.path.join(outdir, "summary.pdf"))

    # Histogram of posterior distributions
    # ----------

    assert os.path.isfile(os.path.join(outdir, "histogram_01.pdf"))

    # Pair plot
    assert os.path.isfile(os.path.join(outdir, "pair_plot.pdf"))



"""Save a text summary comparing parameters from different models"""

from tarpan.shared.info_path import InfoPath
from tarpan.shared.summary import SummaryParams
from tarpan.shared.param_names import filter_param_names

from tarpan.shared.compare_parameters import (
    CompareParametersType,
    save_compare_parameters as shared_save_compare_parameters)


def save_compare_parameters(
        fits,
        labels,
        extra_values=[],
        param_names=None,
        type: CompareParametersType = CompareParametersType.TEXT,
        info_path=InfoPath(),
        summary_params=SummaryParams()):
    """
    Saves a text table that compares model parameters

    Parameters
    ----------

    fits : list of cmdstanpy.stanfit.CmdStanMCMC

        Contains the samples from cmdstanpy.

    labels : list of str

        Names of the models in `fits` list.

    extra_values : list of dict
        Additional values to be shown in the table:

        [
            {
                "mu": 2.3,
                "sigma": 3.3
            }
        ]

    param_names : list of str

        Names of parameters. Include all if None.

    type : CompareParametersType

        Format of values in the text table.

    info_path : InfoPath

        Path information for creating summaries.

    """

    info_path.set_codefile()
    models = []

    for fit in fits:
        param_names = filter_param_names(fit.column_names, param_names)
        samples = fit.get_drawset(params=param_names)
        models.append(samples)

    shared_save_compare_parameters(models, labels=labels,
                                   extra_values=extra_values,
                                   type=type,
                                   param_names=param_names,
                                   info_path=info_path,
                                   summary_params=summary_params)



"""Contains main function that does all analysis"""

from tarpan.shared.info_path import InfoPath
from tarpan.shared.summary import SummaryParams
from tarpan.shared.tree_plot import make_tree_plot
from tarpan.shared.histogram import save_histogram_from_summary
from tarpan.shared.pair_plot import save_pair_plot
from tarpan.shared.summary import save_summary


def save_analysis(samples, param_names=None, info_path=InfoPath(),
                  summary_params=SummaryParams()):
    """
    Creates all analysis files at once: summary, trace and posterior.

    Parameters
    -----------

    samples : Panda's DataFrame

        Each column contains samples from posterior distribution.

    param_names : list of str

        Names of parameters to plot. Plot all parameters if None.
    """

    info_path.set_codefile()

    summary = save_summary(
        samples, param_names=param_names, info_path=info_path,
        summary_params=summary_params)

    make_tree_plot(summary['df'], param_names=param_names, info_path=info_path,
                   summary_params=summary_params)

    save_histogram_from_summary(
        samples, summary['df'], param_names=param_names,
        info_path=info_path, summary_params=summary_params)

    save_pair_plot(samples, param_names=param_names, info_path=info_path)



import shutil
import os
import pandas as pd
from tarpan.shared.pair_plot import save_pair_plot


def test_save_tree_plot():
    data = {
        "x": [1, 2, 2.1, 4, 5, 6, 3.2, 3, 4.4, 2],
        "y": [-1, -2.5, -2, -2.7, -5, -6, -2, -6, -3, -2.5]
    }

    df = pd.DataFrame(data)

    outdir = "tarpan/shared/model_info/pair_plot_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_pair_plot(df, param_names=["x", "y"])

    assert os.path.isfile(os.path.join(outdir, "pair_plot.pdf"))



from dataclasses import dataclass, field
from typing import List
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
from itertools import cycle
from tarpan.shared.info_path import InfoPath, get_info_path
from tarpan.plot.utils import remove_ticks_labels


@dataclass
class ScatterKdeParams:
    title: str = None  # Plot's title
    xlabel: str = None
    ylabel1: str = None
    ylabel2: str = None
    legend_labels = None  # Labels for the plot legend. No legend if None

    marker_colors: List = field(
        default_factory=lambda:
        ["#0060ff44", '#ff002144', '#8888FF44', '#BBBB1144'])

    marker_edgecolors: List = field(
        default_factory=lambda: ['#0060ff', '#ff0021', '#8888FF', '#BBBB11'])

    errorbar_colors: List = field(
        default_factory=lambda:
        ["#0060ff66", '#ff002144', '#8888FF44', '#BBBB1144'])

    markers: List = field(
        default_factory=lambda: ["^", "o", 'x', '*'])

    kde_facecolors: List = field(
        default_factory=lambda:
        ["#0060ff44", '#ff002144', '#8888FF44', '#BBBB1144'])

    kde_edgecolors: List = field(
        default_factory=lambda: ["#0060ff", '#ff0021', '#8888FF', '#BBBB11'])

    kde_show_shade: bool = True  # Show shade under KDE plot
    kde_line_width: float = 1    # Line width of the KDE curve

    grid_color: str = "#aaaaaa"
    grid_alpha: float = 0.2
    markersize: float = 80

    plot_width: float = 6
    plot_height: float = 6


def save_scatter_and_kde(values,
                         uncertainties,
                         title=None,
                         xlabel=None,
                         ylabel=None,
                         info_path=InfoPath(),
                         scatter_kde_params=ScatterKdeParams(),
                         legend_labels=None,
                         plot_fn=None):
    """
    Create a scatter plot and a KDE plot under it.
    The plot is saved to a file.
    The KDE plot uses uncertainties of each individual observation.

    Parameters
    ----------
    values: list of lists
        List of values to plot. Supply more than one list to
        see distributions shown with different colors and markers.
    uncertainties: list of lists
        Uncertainties coresponding to the `values`.
    plot_fn: [function(fig, axes, params), params]
        function:
            A function that can be used to add extra information to the
            plot before it is saved.

            Parameters
            ----------

            fig: Matplotlib's figure object
            axes: list of Matplotlib's axes objects
            params: custom parameters that are passed to the function
        params: custom parameters that will be passed to the function

    Returns
    --------
    fig : Matplotlib's figure object
    axes : list of Matplotlib's axes
    """

    fig, axes = scatter_and_kde(values=values,
                                uncertainties=uncertainties,
                                title=title,
                                xlabel=xlabel,
                                ylabel=ylabel,
                                scatter_kde_params=scatter_kde_params,
                                legend_labels=legend_labels)

    if plot_fn is not None:
        plot_fn[0](fig, axes, params=plot_fn[1])

    info_path.set_codefile()
    info_path.base_name = info_path.base_name or "scatter_kde"
    info_path.extension = info_path.extension or 'pdf'
    plot_path = get_info_path(info_path)
    fig.savefig(plot_path, dpi=info_path.dpi)

    return fig, axes


def scatter_and_kde(values,
                    uncertainties,
                    title=None,
                    xlabel=None,
                    ylabel=None,
                    scatter_kde_params=ScatterKdeParams(),
                    legend_labels=None):
    """
    Create a scatter plot and a KDE plot under it.
    The KDE plot uses uncertainties of each individual observation.

    Parameters
    ----------
    values: list of lists
        List of values to plot. Supply more than one list to
        see distributions shown with different colors and markers.
    uncertainties: list of lists
        Uncertainties coresponding to the `values`.

    Returns
    --------
    fig : Matplotlib's figure object
    axes : list of Matplotlib's axes
    """

    if title is not None:
        scatter_kde_params.title = title

    if xlabel is not None:
        scatter_kde_params.xlabel = xlabel

    if ylabel is not None:
        if isinstance(ylabel, list):
            if len(ylabel) == 2:
                scatter_kde_params.ylabel1 = ylabel[0]
                scatter_kde_params.ylabel2 = ylabel[1]
            elif len(ylabel) == 1:
                scatter_kde_params.ylabel1 = ylabel[0]
        else:
            scatter_kde_params.ylabel1 = ylabel

    if legend_labels is not None:
        scatter_kde_params.legend_labels = legend_labels

    sns.set(style="ticks")

    fig, (ax1, ax2) = plt.subplots(2, 1, sharex=True,
                                   figsize=(scatter_kde_params.plot_width,
                                            scatter_kde_params.plot_height),
                                   gridspec_kw={'hspace': 0})

    # Make a scatter plot
    # ---------------

    marker_colors = cycle(scatter_kde_params.marker_colors)
    marker_edgecolors = cycle(scatter_kde_params.marker_edgecolors)
    errorbar_colors = cycle(scatter_kde_params.errorbar_colors)
    markers = cycle(scatter_kde_params.markers)

    for i, (values_list, uncertainties_list) in enumerate(zip(values, uncertainties)):
        marker_color = next(marker_colors)
        marker_edgecolor = next(marker_edgecolors)
        errorbar_color = next(errorbar_colors)
        marker = next(markers)

        ax1.errorbar(values_list, range(len(values_list)),
                     xerr=uncertainties_list, fmt='none',
                     ecolor=errorbar_color,
                     elinewidth=1,
                     zorder=1)

        value_label = '_nolegend_'

        if scatter_kde_params.legend_labels is not None:
            value_label = scatter_kde_params.legend_labels[i]

        ax1.scatter(values_list, range(len(values_list)),
                    marker=marker,
                    color=marker_color,
                    edgecolor=marker_edgecolor,
                    s=scatter_kde_params.markersize,
                    zorder=2,
                    label=value_label)

    if scatter_kde_params.ylabel1 is not None:
        ax1.set_ylabel(scatter_kde_params.ylabel1)

    if scatter_kde_params.legend_labels is not None:
        ax1.legend()

    # Make a density plot
    # -----------------

    xlims = ax1.get_xlim()
    pad = abs(xlims[0] - xlims[1]) / 10
    x = np.linspace(xlims[0] - pad, xlims[1] + pad, 1000)

    kde_facecolors = cycle(scatter_kde_params.kde_facecolors)
    kde_edgecolors = cycle(scatter_kde_params.kde_edgecolors)

    for values_list, uncertainties_list in zip(values, uncertainties):
        kde_facecolor = next(kde_facecolors)
        kde_edgecolor = next(kde_edgecolors)

        y = gaussian_kde(x, values_list, uncertainties_list)

        if scatter_kde_params.kde_show_shade:
            ax2.fill_between(x, y1=y,
                             edgecolor=None,
                             facecolor=kde_facecolor,
                             linewidth=0)

        ax2.plot(x, y, c=kde_edgecolor,
                 linewidth=scatter_kde_params.kde_line_width,
                 zorder=5)

    if scatter_kde_params.xlabel is not None:
        ax2.set_xlabel(scatter_kde_params.xlabel)

    if scatter_kde_params.ylabel2 is not None:
        ax2.set_ylabel(scatter_kde_params.ylabel2)

    if scatter_kde_params.title is not None:
        fig.suptitle(scatter_kde_params.title)

    ax1.grid(color=scatter_kde_params.grid_color, linewidth=1,
             alpha=scatter_kde_params.grid_alpha)

    ax2.grid(color=scatter_kde_params.grid_color, linewidth=1,
             alpha=scatter_kde_params.grid_alpha)

    remove_ticks_labels(ax1, remove_x=False, remove_y=True)
    remove_ticks_labels(ax2, remove_x=False, remove_y=True)

    if scatter_kde_params.title is not None:
        fig.tight_layout(rect=[0, 0, 1, 0.95])
    else:
        fig.tight_layout()

    return fig, [ax1, ax2]


def gaussian_kde(eval_points, values, uncertainties):
    """
    Computer Gaussian KDE (kernel density estimator) using values and their
    uncertianties by adding PDFs of Normal distributions at each value that
    have mean equal to the value and standard deviation equal to value's
    uncertainty.

    Parameters
    ----------

    eval_points: list or numpy.ndarray
        Value at which KDE will be evaluated.
    values: list or numpy.ndarray
        Values for computing KDE.
    uncertainties: list or numpy.ndarray
        Corresponding uncertainties for the `values`.

    Returns
    -------
    numpy.ndarray
        Values of the KDE corresponding to `eval_points`.
    """

    x = np.array(eval_points)
    mu = np.array(values)
    sigma = np.array(uncertainties)

    if mu.shape[0] == 0:
        return np.array([])

    if mu.shape[0] != sigma.shape[0]:
        raise ValueError('Values and uncertainties\
are lists of different lengths')

    guassian = -0.5 * ((x[:, np.newaxis] - mu[np.newaxis, :]) / sigma)**2
    gaussian = np.exp(guassian)
    gaussian /= np.sqrt(2 * np.pi) * sigma
    return gaussian.sum(axis=1) / len(values)



import os
import shutil
from tarpan.testutils.a01_eight_schools.eight_schools import get_fit
from tarpan.cmdstanpy.traceplot import save_traceplot


def test_save_traceplot():
    fit = get_fit()
    outdir = "tarpan/cmdstanpy/model_info/traceplot_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_traceplot(fit)

    assert os.path.isfile(os.path.join(outdir, "traceplot_01.pdf"))
    assert os.path.isfile(os.path.join(outdir, "traceplot_02.pdf"))
    assert os.path.isfile(os.path.join(outdir, "traceplot_03.pdf"))



import os
import shutil
from tarpan.cmdstanpy.summary import save_summary
from tarpan.testutils.a01_eight_schools.eight_schools import get_fit
from tarpan.shared.summary import SummaryParams


def test_save_summary():
    fit = get_fit()

    outdir = "tarpan/cmdstanpy/model_info/summary_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_summary(fit, param_names=["mu", "tau", 'eta.1'])

    assert os.path.isfile(os.path.join(outdir, "summary.txt"))
    assert os.path.isfile(os.path.join(outdir, "summary.csv"))


def test_save_summary_specify_hpdi():
    fit = get_fit()

    outdir = "tarpan/cmdstanpy/model_info/summary_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_summary(fit, param_names=["mu", "tau", 'eta.1'],
                 summary_params=SummaryParams(hpdis=[0.05, 0.99]))

    assert os.path.isfile(os.path.join(outdir, "summary.txt"))
    assert os.path.isfile(os.path.join(outdir, "summary.csv"))



import os
import shutil

from tarpan.cmdstanpy.compare import save_compare

from tarpan.testutils.a05_divorse.divorse import (
    get_fit1_divorse_age, get_fit2_divorse_marriage,
    get_fit3_divorse_age_marriage)


def test_save_compare():
    fit1_divorse_age = get_fit1_divorse_age()
    fit2_divorse_marriage = get_fit2_divorse_marriage()
    fit3_divorse_age_marriage = get_fit3_divorse_age_marriage()

    models = {
        "Divorse vs Age": fit1_divorse_age,
        "Divorse vs Marriage": fit2_divorse_marriage,
        "Divorse vs Age+Marriage": fit3_divorse_age_marriage
    }

    outdir = "tarpan/cmdstanpy/model_info/compare_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_compare(models=models)

    # WAIC
    # ------

    assert os.path.isfile(os.path.join(outdir, "compare_waic.csv"))
    assert os.path.isfile(os.path.join(outdir, "compare_waic.txt"))
    assert os.path.isfile(os.path.join(outdir, "compare_waic.pdf"))

    # PSIS
    # ------

    assert os.path.isfile(os.path.join(outdir, "compare_psis.csv"))
    assert os.path.isfile(os.path.join(outdir, "compare_psis.txt"))
    assert os.path.isfile(os.path.join(outdir, "compare_psis.pdf"))

    assert os.path.isfile(os.path.join(outdir,
                          "pareto_k_divorse_vs_age_marriage.pdf"))

    assert os.path.isfile(os.path.join(outdir,
                                       "pareto_k_divorse_vs_age.pdf"))

    assert os.path.isfile(os.path.join(outdir,
                                       "pareto_k_divorse_vs_marriage.pdf"))



import os
import pickle
from tarpan.shared.info_path import InfoPath, get_info_dir


def run(func, info_path=InfoPath(), **kwargs):
    """
    A function that helps to run Stan only once and return cached result
    next time instead.

    Parameters
    ----------
    fun: function
        A function that initialises CmdStanModel() model, runs
        model.sample(...) and returns its result (cmdstanpy.CmdStanMCMC object).
        The function must have an `output_dir` parameter, which should be
        pass to model.sample:
            model.sample(output_dir=output_dir)
        for storing the stan's CSV files in correct location.

    **kwargs: any other argument that will be passed to `func`

    Returns
    -------
    mdstanpy.CmdStanMCMC
        Contains the samples from cmdstanpy.
    """

    info_path.set_codefile()
    fit_dir = os.path.join(get_info_dir(info_path), "stan_cache")
    fit_path = os.path.join(fit_dir, "fit.pkl")

    if not os.path.exists(fit_path):
        os.makedirs(fit_dir, exist_ok=True)
        fit = func(**kwargs, output_dir=fit_dir)

        # Load samples into the fit object, so they can be retrived faster
        temp = fit.get_drawset()

        with open(fit_path, 'wb') as file:
            pickle.dump(fit, file, protocol=pickle.HIGHEST_PROTOCOL)

    with open(fit_path, 'rb') as input:
        return pickle.load(input)



import shutil
import os

from tarpan.testutils.a01_eight_schools.eight_schools import (
    get_fit, get_fit_larger_uncertainties)

from tarpan.cmdstanpy.compare_parameters import save_compare_parameters


def test_save_compare_parameters():
    fit = get_fit()
    fit2 = get_fit_larger_uncertainties()
    fits = [fit, fit2]

    outdir = "tarpan/cmdstanpy/model_info/compare_parameters_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_compare_parameters(fits, labels=['One', 'Two'],
                            param_names=["mu", "tau", "eta.1"])

    assert os.path.isfile(os.path.join(outdir, "parameters_compared.txt"))

    with open(os.path.join(outdir, "parameters_compared.txt"), 'r') as file:
        data = file.read()
        assert "2.36 (+5.41, -2.35)" in data


def test_save_compare_parameters__extra_values():
    fit = get_fit()
    fit2 = get_fit_larger_uncertainties()
    fits = [fit, fit2]

    outdir = "tarpan/cmdstanpy/model_info/compare_parameters_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    extra_values = [
        {
            "mu": 1.1,
            "tau": 2.1,
            "eta.1": 3.1,
            "eta.2": 5.2
        },
        {
            "mu": 10.1,
            "tau": 20.1,
            "eta.1": 30.1,
            "eta.2": 35.3
        }
    ]

    save_compare_parameters(fits, labels=['One', 'Two', 'Extra 1', 'Extra 2'],
                            extra_values=extra_values,
                            param_names=["mu", "tau", "eta.1"])

    assert os.path.isfile(os.path.join(outdir, "parameters_compared.txt"))

    with open(os.path.join(outdir, "parameters_compared.txt"), 'r') as file:
        data = file.read()
        assert "2.36 (+5.41, -2.35)" in data
        assert "Extra 1" in data
        assert "Extra 2" in data
        assert "30.1" in data



import os
from tarpan.shared.info_path import InfoPath, get_info_path, get_info_dir


def test_set_codefile():
    info_path = InfoPath()

    set_codefile_sun_function(info_path)

    assert os.path.basename(info_path.codefile_path) == "info_path_test.py"


def set_codefile_sun_function(info_path):
    info_path.set_codefile()


def test_get_info_path():
    info_path = InfoPath()
    info_path.base_name = 'my_basename'
    info_path.extension = 'test_extension'
    result = get_info_path(info_path)

    assert 'tarpan/shared/model_info/info_path_test/my_basename\
.test_extension' in result


def test_get_info_dir():
    info_path = InfoPath()
    result = get_info_dir(info_path)

    assert 'tarpan/shared/model_info/info_path_test' in result



import os
import scipy.stats as stats
import shutil
from tarpan.plot.posterior import save_posterior_scatter_and_kde

from tarpan.testutils.a02_gaussian_mixture.gaussian_mixture import (
    get_fit1, get_fit2, get_data1, get_data2)


def model_pdf(x, row):
    mu1 = row['mu.1']
    mu2 = row['mu.2']
    sigma = row['sigma']
    r = row['r']

    return (1 - r) * stats.norm.pdf(x, mu1, sigma) + \
        r * stats.norm.pdf(x, mu2, sigma)


def test_save_posterior_scatter_and_kde():
    fit1 = get_fit1()
    fit2 = get_fit2()
    fits = [fit1, fit2]

    outdir = "tarpan/plot/model_info/posterior_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    data1 = get_data1()
    data2 = get_data2()

    fig, axes = save_posterior_scatter_and_kde(
        fits=fits,
        pdf=model_pdf,
        values=[data1["y"], data2["y"]],
        uncertainties=[data1["uncertainties"], data2["uncertainties"]],
        title="Sodium abundances in RGB stars of NGC 288",
        xlabel="Sodium abundance [Na/H]",
        ylabel=["Star number", "Probability density"],
        legend_labels=["AGB", "RGB"])

    assert os.path.isfile(os.path.join(outdir, "posterior_scatter_kde.pdf"))
    assert len(axes) == 2



"""Make histograms of parameter distributions"""

from tarpan.shared.info_path import InfoPath

from tarpan.shared.histogram import (
    HistogramParams, save_histogram_from_summary)

from tarpan.cmdstanpy.summary import make_summary
from tarpan.shared.summary import SummaryParams


def save_histogram(fit,
                   param_names=None,
                   info_path=InfoPath(),
                   summary_params=SummaryParams(),
                   histogram_params=HistogramParams()):
    """
    Make histograms of parameter distributions.

    Parameters
    ----------

    fit : cmdstanpy.stanfit.CmdStanMCMC
        Samples from cmdstanpy.

    param_names : list of str
        Names of parameters to be included in the summar. Include all if None.

    info_path : InfoPath
        Path information for creating summaries.

    """

    info_path.set_codefile()

    df_summary, summary, samples = make_summary(
        fit, param_names=param_names, summary_params=summary_params)

    save_histogram_from_summary(
        samples, df_summary, param_names=param_names,
        info_path=info_path, summary_params=summary_params,
        histogram_params=histogram_params)



import shutil
import os
import pandas as pd
from tarpan.shared.tree_plot import TreePlotParams, save_tree_plot


def test_save_tree_plot():
    data = {
        "x": [1, 2, 3, 4, 5, 6],
        "y": [-1, -2, -3, -4, -5, -6]
    }

    df = pd.DataFrame(data)

    values_no_error_bars = [
        {
            "x": 1.1,
            "y": -3.1,
        }
    ]

    tree_params = TreePlotParams()

    tree_params.labels = [
        "Normal",
        "Exact",
    ]

    outdir = "tarpan/shared/model_info/tree_plot_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_tree_plot(
        [df], extra_values=values_no_error_bars, param_names=["mu", "tau"],
        tree_params=tree_params)

    assert os.path.isfile(os.path.join(outdir, "summary.pdf"))



from tarpan.cmdstanpy.psis import (
    save_compare_psis_csv_from_compared,
    save_compare_psis_txt_from_compared,
    save_compare_psis_tree_plot_from_compared,
    compare_psis, save_psis_pareto_k_plot_from_compared,
    ParetoKPlotParams)

from tarpan.cmdstanpy.waic import (
    save_compare_waic_csv_from_compared,
    save_compare_waic_txt_from_compared,
    save_compare_waic_tree_plot_from_compared,
    compare_waic)

from tarpan.cmdstanpy.waic import LPD_COLUMN_NAME_DEFAULT
from tarpan.shared.tree_plot import TreePlotParams
from tarpan.shared.info_path import InfoPath


def save_compare(
        models, lpd_column_name=LPD_COLUMN_NAME_DEFAULT,
        tree_plot_params: TreePlotParams = TreePlotParams(),
        info_path=InfoPath(),
        pareto_k_plot_params: ParetoKPlotParams = ParetoKPlotParams()):
    """
    Compare multiple models using WAIC (Widely Aplicable Information Criterion)
    and PSIS (Pareto-smoothed importance sampling) methods. Saves the analysis
    data and plots.

    Parameters
    ----------

    models : list of dict
        List of model samples from cmdstanpy to compare.

        The dictionary has keys:
            name: str
                Model name
            fit: cmdstanpy.stanfit.CmdStanMCMC
                Contains the samples from cmdstanpy.

    lpd_column_name : str
        Prefix of the columns in Stan's output that contain log
        probability density value for each observation. For example,
        if lpd_column_name='possum', when output is expected to have
        columns 'possum.1', 'possum.2', ..., 'possum.33' given 33 observations.


    info_path : InfoPath
        Determines the location of the output files.
    """

    info_path.set_codefile()

    # Compare with WAIC
    # --------

    compared = compare_waic(models=models, lpd_column_name=lpd_column_name)
    save_compare_waic_csv_from_compared(compared=compared, info_path=info_path)
    save_compare_waic_txt_from_compared(compared=compared, info_path=info_path)

    save_compare_waic_tree_plot_from_compared(
        compared=compared,
        tree_plot_params=tree_plot_params,
        info_path=info_path)

    # Compare with PSIS
    # --------

    compared = compare_psis(models=models, lpd_column_name=lpd_column_name)
    save_compare_psis_csv_from_compared(compared=compared, info_path=info_path)
    save_compare_psis_txt_from_compared(compared=compared, info_path=info_path)

    save_compare_psis_tree_plot_from_compared(
        compared=compared,
        tree_plot_params=tree_plot_params,
        info_path=info_path)

    save_psis_pareto_k_plot_from_compared(
        compared=compared,
        info_path=info_path,
        pareto_k_plot_params=pareto_k_plot_params)



import numpy as np


def hpdi(values, probability):
    """
    Computes HPDI (Highest Posterior Density Interval),
    which is the inteval of minimum width that includes the given
    `probability` (or proportion of the numbers)

    The following code is based on from `hpd` function from
    arviz library (https://github.com/arviz-devs/arviz), version 0.6.1,
    distributed under Apache License Version 2.0.

    Changes made to arviz code:
     * Kept only code for single, non-circular interval.

    Parameters
    ----------

    values: list of float
        List of numbers.
    probability: float
        Probability (or compatibility/credible interval), or fraction
        of numbers.

    Returns
    -------
    Tuple containing lower and upper boundaries of the interval.
    """

    values = np.sort(values)
    n = len(values)
    interval_idx_inc = int(np.floor(probability * n))
    n_intervals = n - interval_idx_inc
    interval_width = values[interval_idx_inc:] - values[:n_intervals]

    if len(interval_width) == 0:
        raise ValueError("Too few elements for interval calculation. ")

    min_idx = np.argmin(interval_width)
    hdi_min = values[min_idx]
    hdi_max = values[min_idx + interval_idx_inc]

    return (hdi_min, hdi_max)



"""Makes a tree plot showing summary of distributions of parameters"""

from dataclasses import dataclass
from tarpan.shared.info_path import InfoPath, get_info_path
import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import math
from itertools import cycle
import re
from tarpan.shared.summary import SummaryParams, sample_summary
from tarpan.shared.param_names import filter_param_names


@dataclass
class TreePlotParams:
    """
    Parameters of the tree plot.
    """

    group_height: float = None
    error_bar_cap_size: float = None
    markersize: float = 50
    ylim: float = None
    markers = ['^', 'o', 'X', 'D', 'X', 'v', '<', '*', '>', '|', 'd']
    marker_line_width: float = 1

    marker_colors = ["#bcd5fd", '#febcc4', '#FF9922',
                     '#009933', '#8888FF',
                     '#BBBB11', '#3399AA', '#33BB11',
                     '#330077', '#BBAA22', '#00AA77',
                     '#EE8844']

    marker_edge_colors = ["#0060ff", '#ff0021', '#994400',
                          '#003300', '#111177',
                          '#888800', '#003355', '#008800',
                          '#000022', '#776600', '#006622',
                          '#662200']

    error_bar_colors = ["#0033BB33", '#7755FF55', '#22FF9955']
    labels = None  # Labels for the plot legend. No legend if None
    figure_width: int = None
    figure_height: int = None
    title: str = None
    xlabel: str = None
    xlim: list = None

    # When true and zero in within x limits of the data, draw the vertical
    # line through zero
    draw_zero_line_if_in_range: bool = True

    grid_color: str = "#aaaaaa"
    grid_alpha: float = 0.2


def make_tree_plot(df_summary, param_names=None, info_path=InfoPath(),
                   tree_params: TreePlotParams = TreePlotParams(),
                   summary_params=SummaryParams()):
    """
    Make tree plot of parameters.
    """

    info_path = InfoPath(**info_path.__dict__)
    tree_plot_data = extract_tree_plot_data(
        df_summary, param_names=param_names, summary_params=summary_params)

    fig, ax = tree_plot(tree_plot_data, params=tree_params)
    info_path.base_name = info_path.base_name or 'summary'
    info_path.extension = info_path.extension or 'pdf'
    the_path = get_info_path(info_path)
    fig.savefig(the_path, dpi=info_path.dpi)
    plt.close(fig)


def summary_from_dict(data):
    """
    Converts values of parmeter to summary data frame, used for
    adding values to tree plot.


    Parameters
    ----------

    data: dict

        Mode values of the parameters, for example

        {
            "a": 12,
            "b": 32,
            "c": 48
        }

    Returns
    -------
    Panda's DataFrame
        Summary dataframe with a single "Mode" column.
    """

    return pd.DataFrame.from_dict(data, orient="index", columns=["Mode"])


def save_tree_plot(models,
                   extra_values=[],
                   param_names=None,
                   info_path=InfoPath(),
                   summary_params=SummaryParams(),
                   tree_params=TreePlotParams()):
    """
    Save a tree plot that summarises parameter distributions.
    Can compare summaries from multiple models, when multiple samples are
    supplied. One can also supply additional markers
    to be compared with using `extra_values` parameter.

    Parameters
    ----------

    models : list Panda's data frames

        List of data frames for each model, containg sample values for
        multiple parameters (one parameter is one data frame column).
        Supply multiple data frames to see their distribution summaries
        compared on the tree plot.

    extra_values : list of dict
        Additional markers to be shown on tree plot, without error bars:

        [
            {
                "mu": 2.3,
                "sigma": 3.3
            }
        ]

    param_names : list of str

        Names of parameters. Include all if None.

    info_path : InfoPath

        Path information for creating summaries.

    """

    info_path.set_codefile()
    summaries = []

    for samples in models:
        column_names = list(samples)
        param_names = filter_param_names(column_names, param_names)
        summary, _ = sample_summary(samples, params=summary_params)
        summaries.append(summary)

    for values in extra_values:
        summaries.append(summary_from_dict(values))

    make_comparative_tree_plot(
        summaries, info_path=info_path,
        tree_params=tree_params)


def make_comparative_tree_plot(
        summaries, param_names=None, info_path=InfoPath(),
        tree_params: TreePlotParams = TreePlotParams()):
    """
    Make tree plot that compares summaries of parameters
    """

    info_path = InfoPath(**info_path.__dict__)
    tree_plot_data = None

    for df_summary in summaries:
        tree_plot_data = extract_tree_plot_data(
            df_summary, param_names=param_names, groups=tree_plot_data)

    fig, ax = tree_plot(tree_plot_data, params=tree_params)
    info_path.base_name = info_path.base_name or 'summary'
    info_path.extension = info_path.extension or 'pdf'
    the_path = get_info_path(info_path)
    fig.savefig(the_path, dpi=info_path.dpi)
    plt.close(fig)


def tree_plot(groups, params: TreePlotParams = TreePlotParams()):
    """
    Make a tree plot of parameters.

    Parameters
    -----------

    groups : list of dict

        Data for plotting. The data structure is described in the
        "Returns" section of the `extract_tree_plot_data` function.

    group_height : float

        The height of each group, a unmber between 0 and 1.
    """

    sns.set(style="ticks")

    if params.error_bar_cap_size is None:
        error_bar_cap_size = len(groups) / 50

    fig, ax = plt.subplots()
    total_markers = 0
    labels_added = []  # Used to avoid duplicate legend labels

    for i_group, group in enumerate(groups):
        elements_in_group = len(group['values'])

        this_group_height = params.group_height

        if this_group_height is None:
            # Determine group height automatically
            x = len(group['values']) - 1
            this_group_height = 1 - math.exp(-x / 5)

        # Vertical separation between values in one group
        if len(group['values']) == 1:
            y_increment = 0
            this_group_height = 0
        else:
            y_increment = 1 / (elements_in_group - 1) * this_group_height

        markers = cycle(params.markers)
        marker_colors = cycle(params.marker_colors)
        marker_edge_colors = cycle(params.marker_edge_colors)

        # Plot the values for the group
        for i_value, value in enumerate(group['values']):
            total_markers += 1
            y_coord = i_group + this_group_height/2 - y_increment * i_value

            value_label = '_nolegend_'

            if params.labels is not None:
                current_label = params.labels[i_value]

                if current_label not in labels_added:
                    value_label = current_label
                    labels_added.append(current_label)

            marker = next(markers)
            marker_color = next(marker_colors)
            marker_edge_color = next(marker_edge_colors)

            ax.scatter(value["value"], y_coord,
                       marker=marker, s=params.markersize,
                       facecolor=marker_color,
                       edgecolor=marker_edge_color,
                       linewidth=params.marker_line_width, zorder=5,
                       label=value_label)

            # Plot error bars
            n_error_bars = len(value["error_bars"])
            error_bar_colors = cycle(params.error_bar_colors)

            for i_error_bar, error_bar in enumerate(value["error_bars"]):
                color = next(error_bar_colors)

                if i_error_bar == n_error_bars - 1:
                    color = marker_edge_color

                ax.plot(error_bar, [y_coord, y_coord], zorder=i_error_bar + 1,
                        color=color)

                # Plot caps
                ax.plot([error_bar[0], error_bar[0]],
                        [y_coord - error_bar_cap_size/2,
                         y_coord + error_bar_cap_size/2],
                        zorder=i_error_bar + 6,
                        color=color)

                ax.plot([error_bar[1], error_bar[1]],
                        [y_coord - error_bar_cap_size/2,
                         y_coord + error_bar_cap_size/2],
                        zorder=i_error_bar + 6,
                        color=color)

    group_names = [group['name'] for group in groups]

    # Set figure size
    # ---------

    if params.figure_width is None or params.figure_height is None:
        # Set width and height automatically based
        params.figure_width = 6

        # Make figure taller if there are more markers to make them spread out
        params.figure_height = 3.5 + total_markers * 0.1

    fig.set_figwidth(params.figure_width)
    fig.set_figheight(params.figure_height)

    # Plot vertical line through zero zero is within the data range
    bounds = sorted([ax.get_xlim()[0], ax.get_xlim()[1]])

    if params.draw_zero_line_if_in_range and\
            bounds[0] < 0 and bounds[1] > 0:
        ax.axvline(x=0, linestyle='dashed', color=params.marker_edge_colors[0])

    if params.labels is not None:
        ax.legend()

    if params.xlim is not None:
        ax.set_xlim(params.xlim)

    if params.xlabel is not None:
        ax.set_xlabel(params.xlabel)

    if params.title is not None:
        ax.set_title(params.title)

    ax.set_yticks(range(0, len(groups)))
    ax.set_yticklabels(group_names)

    ax.grid(axis='x', zorder=-10,
            color=params.grid_color, linewidth=1,
            alpha=params.grid_alpha)

    fig.tight_layout()

    if params.ylim is not None:
        ax.set_ylim(params.ylim)

    return (fig, ax)


def extract_tree_plot_data(df, param_names=None, groups=None,
                           summary_params=SummaryParams()):
    """
    Extract data used to for tree plot function from a dataframe.

    Parameters
    -----------

    param_names: list of str

        List of parameters to plot. If None, plot all.

    df : Panda's data frame
        Data frame containing summary

    groups : list
        Tree plot data. If passed, the data frame's data will be added to it.

    Returns
    -------

    Array of dictionaries that will be used to make tree plot.

    For example, here we plot values of two variables "temperature"
    and "pressure" from two observations.
    Each value has multiple error bars, the 95% and 68% bars, for example.

    [
        {
            "name": "temperature"
            "values": [
                {
                    value: 10,
                    error_bars: [[6, 16], [9, 11]]
                },
                {
                    value: 40,
                    error_bars: [[10, 80], [38, 42]]
                },
            ]
        },
        {
            "name": "pressure"
            "values": [
                {
                    value: 1.1,
                    error_bars: [[0.1, 2.3], [0.9, 1.2]]
                },
                {
                    value: 1.6,
                    error_bars: [[0.6, 2.7], [1.1, 1.9]]
                }
            ]
        }
    ]
    """

    if groups is None:
        groups = []

    for column_name, row in df.iterrows():
        if param_names is not None:
            # If param_names contains 'a', we will also plot
            # parameters named 'a.1', 'a.2' etc.
            if (column_name not in param_names
               and re.sub(r'\.[0-9]+\Z', '', column_name) not in param_names):
                continue

        column_summary = row
        group_data = None

        # Check if `groups` already has the column
        for group in groups:
            if group['name'] == column_name:
                group_data = group
                break

        # Have not found group data - create one
        if group_data is None:
            group_data = {
                'name': column_name,
                'values': []
            }

            groups.append(group_data)

        # Add new value

        value = {}
        value['value'] = column_summary["Mode"]
        group_data['values'].append(value)

        # Add error bars from the HPDI values
        # ---------

        error_bars = []
        hpdis = sorted(summary_params.hpdi_percent(), reverse=True)

        for hpdi in hpdis:
            start = f'{hpdi}CI-'
            end = f'{hpdi}CI+'

            if start in column_summary:
                error_bars.append([
                    column_summary[start],
                    column_summary[end]])

        value['error_bars'] = error_bars

    return groups



import shutil
import os
import pandas as pd

from tarpan.shared.compare_parameters import (
    save_compare_parameters, format_parameter, format_value,
    CompareParametersType)


def test_format_parameter__text():
    data = {
        "Mode": 1.23,
        "+": 0.02,
        "-": 0.06
    }

    result = format_parameter(data, CompareParametersType.TEXT)

    assert result == "1.23 (+0.02, -0.06)"


def test_format_parameter__gitlab_latex():
    data = {
        "Mode": 1.23,
        "+": 0.02,
        "-": 0.06
    }

    result = format_parameter(data, CompareParametersType.GITLAB_LATEX)

    assert result == "$`1.23^{+0.02}_{-0.06}`$"


def test_format_value__text():
    result = format_value(1.23, CompareParametersType.TEXT)

    assert result == "1.23"


def test_format_value__gitlab_latex():
    result = format_value(1.23, CompareParametersType.GITLAB_LATEX)

    assert result == "$`1.23`$"


def test_save_tree_plot():
    data1 = {
        "x": [1, 2, 3, 4, 5, 6],
        "y": [-1, -2, -3, -4, -5, -6],
        "z": [40, 21, 32, 41, 11, 31]
    }

    df1 = pd.DataFrame(data1)

    data2 = {
        "x": [2, 3, 1, 1, 3, 4],
        "y": [-2.1, -2, -2, -3, -1, -4],
        "z": [23, 19, 21, 13, 29, 10]
    }

    df2 = pd.DataFrame(data2)

    outdir = "tarpan/shared/model_info/compare_parameters_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_compare_parameters([df1, df2], labels=['One', 'Two'],
                            param_names=["x", "y"])

    assert os.path.isfile(os.path.join(outdir, "parameters_compared.txt"))

    with open(os.path.join(outdir, "parameters_compared.txt"), 'r') as file:
        data = file.read()
        assert "-2.10 (+0.10, -1.90)" in data


def test_save_tree_plot__gitlab_latex():
    data1 = {
        "x": [1, 2, 3, 4, 5, 6],
        "y": [-1, -2, -3, -4, -5, -6],
        "z": [40, 21, 32, 41, 11, 31]
    }

    df1 = pd.DataFrame(data1)

    data2 = {
        "x": [2, 3, 1, 1, 3, 4],
        "y": [-2.1, -2, -2, -3, -1, -4],
        "z": [23, 19, 21, 13, 29, 10]
    }

    df2 = pd.DataFrame(data2)

    outdir = "tarpan/shared/model_info/compare_parameters_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_compare_parameters([df1, df2], labels=['One', 'Two'],
                            param_names=["x", "y"],
                            type=CompareParametersType.GITLAB_LATEX)

    assert os.path.isfile(os.path.join(outdir, "parameters_compared.txt"))

    with open(os.path.join(outdir, "parameters_compared.txt"), 'r') as file:
        data = file.read()
        assert "$`-2.10^{+0.10}_{-1.90}`$" in data


def test_save_tree_plot__extra_values():
    data1 = {
        "x": [1, 2, 3, 4, 5, 6],
        "y": [-1, -2, -3, -4, -5, -6],
        "z": [40, 21, 32, 41, 11, 31]
    }

    df1 = pd.DataFrame(data1)

    data2 = {
        "x": [2, 3, 1, 1, 3, 4],
        "y": [-2.1, -2, -2, -3, -1, -4],
        "z": [23, 19, 21, 13, 29, 10]
    }

    df2 = pd.DataFrame(data2)

    outdir = "tarpan/shared/model_info/compare_parameters_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    extra_values = [
        {
            "x": 1.1,
            "y": 2.1,
            "z": 3.1
        },
        {
            "x": 10.1,
            "y": 20.1,
            "z": 30.1
        }
    ]

    save_compare_parameters([df1, df2],
                            labels=['One', 'Two', 'Extra 1', 'Extra 2'],
                            extra_values=extra_values,
                            param_names=["x", "y"])

    assert os.path.isfile(os.path.join(outdir, "parameters_compared.txt"))

    with open(os.path.join(outdir, "parameters_compared.txt"), 'r') as file:
        data = file.read()
        assert "Extra 1" in data
        assert "-2.10 (+0.10, -1.90)" in data
        assert "1.10" in data


def test_save_tree_plot__extra_values_missing():
    data1 = {
        "x": [1, 2, 3, 4, 5, 6],
        "y": [-1, -2, -3, -4, -5, -6],
        "z": [40, 21, 32, 41, 11, 31]
    }

    df1 = pd.DataFrame(data1)

    data2 = {
        "x": [2, 3, 1, 1, 3, 4],
        "y": [-2.1, -2, -2, -3, -1, -4],
        "z": [23, 19, 21, 13, 29, 10]
    }

    df2 = pd.DataFrame(data2)

    outdir = "tarpan/shared/model_info/compare_parameters_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    extra_values = [
        {
            "x": 1.1
        },
        {
            "x": 10.1
        }
    ]

    save_compare_parameters([df1, df2],
                            labels=['One', 'Two', 'Extra 1', 'Extra 2'],
                            extra_values=extra_values,
                            param_names=["x", "y"])

    assert os.path.isfile(os.path.join(outdir, "parameters_compared.txt"))

    with open(os.path.join(outdir, "parameters_compared.txt"), 'r') as file:
        data = file.read()
        assert "Extra 1" in data
        assert "-2.10 (+0.10, -1.90)" in data
        assert "1.10" in data



"""Functions for dealing with plots"""

import numpy as np
import seaborn as sns


def remove_ticks_labels(ax, remove_x=True, remove_y=True):
    """Remove ticks and labels from axes"""

    if remove_x:
        ax.set_xticklabels([])
        ax.set_xticks([])

    if remove_y:
        ax.set_yticklabels([])
        ax.set_yticks([])


def plot_kde_fallback_hist(samples, **kwargs):
    """
    Plot KDE of the sample and fall back to histogram if KDE fails
    """

    try:
        sns.kdeplot(samples, **kwargs)
    except np.linalg.LinAlgError:
        sns.distplot(samples, norm_hist=True, kde=False, **kwargs)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="tarpan",
    version="0.3.8",
    author="Evgenii Neumerzhitckii",
    author_email="sausageskin@gmail.com",
    description="A collection of tools for analysing cmdspanpy output, written in Python",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/evgenyneu/tarpan",
    packages=setuptools.find_packages(),
    install_requires=[
        'cmdstanpy>=0.8.0,<0.9.0',
        'scipy>=1.4.1,<2.0.0',
        'tabulate>=0.8.6,<0.9.0',
        'pandas>=0.25.3,<0.26.0',
        'numpy>=1.18.1,<2.0.0',
        'matplotlib>=3.1.2,<4.0.0',
        'seaborn>=0.10.0,<0.11.0'
    ],
    tests_require=[
        'pytest==5.3.0'
    ],
    license='Public Domain',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Public Domain",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)



"""Makes a tree plot showing summary of distributions of parameters"""

from tarpan.shared.info_path import InfoPath
from tarpan.shared.tree_plot import (
    TreePlotParams, summary_from_dict, make_comparative_tree_plot)

from tarpan.shared.summary import (
    SummaryParams, sample_summary)

from tarpan.shared.param_names import filter_param_names


def save_tree_plot(fits,
                   extra_values=[],
                   param_names=None,
                   info_path=InfoPath(),
                   summary_params=SummaryParams(),
                   tree_params=TreePlotParams()):
    """
    Save a tree plot that summarises parameter distributions.
    Can compare summaries from multiple models, when multiple fits are
    supplied. One can also supply additional markers
    to be compared with using `extra_values` parameter.

    Parameters
    ----------

    fits : list of cmdstanpy.stanfit.CmdStanMCMC

        Contains the samples from cmdstanpy.

    extra_values : list of dict
        Additional markers to be shown on tree plot, without error bars:

        [
            {
                "mu": 2.3,
                "sigma": 3.3
            }
        ]

    param_names : list of str

        Names of parameters. Include all if None.

    info_path : InfoPath

        Path information for creating summaries.

    """

    info_path.set_codefile()
    summaries = []

    for fit in fits:
        param_names = filter_param_names(fit.column_names, param_names)
        samples = fit.get_drawset(params=param_names)
        summary, _ = sample_summary(samples, params=summary_params)
        summaries.append(summary)

    for values in extra_values:
        summaries.append(summary_from_dict(values))

    make_comparative_tree_plot(
        summaries, info_path=info_path,
        tree_params=tree_params)



"""Save a text summary comparing parameters from different models"""

import pandas as pd
from tabulate import tabulate
from enum import Enum
from tarpan.shared.info_path import InfoPath, get_info_path
from tarpan.shared.summary import SummaryParams, sample_summary
from tarpan.shared.param_names import filter_param_names


class CompareParametersType(Enum):
    """Style of presenting values"""

    """Value and uncertainty is presented as 1.23 (+0.02, -0.03)"""
    TEXT = 1

    """
    Value and uncertainty is presented in gitlab latex format:
        $`1.23^{0.02}_{0.03}`$
    """
    GITLAB_LATEX = 2


def format_value(value, type: CompareParametersType):
    """
    Converts value to text

    Parameters
    ----------

    value: float
        A value to be converte to text.

    type:
        Style of prepresenting a number in text
    """

    if type == CompareParametersType.TEXT:
        return f"{value:.2f}"
    else:
        value_txt = f"{value:.2f}"
        return f"$`{value_txt}`$"


def format_parameter(data, type: CompareParametersType):
    """
    Converts a parameter value to text

    Parameters
    ----------

    data:
        a dictionary or Panda's row containing parameter values

    type:
        Style of prepresenting a number in text
    """

    if type == CompareParametersType.TEXT:
        return f"{data['Mode']:.2f} (+{data['+']:.2f}, -{data['-']:.2f})"
    else:
        value = f"{data['Mode']:.2f}"
        upper = f"{data['+']:.2f}"
        lower = f"{data['-']:.2f}"

        return f"$`{value}^{{+{upper}}}_{{-{lower}}}`$"


def compare_parameters(
        models,
        labels,
        extra_values=[],
        type: CompareParametersType = CompareParametersType.TEXT,
        param_names=None,
        summary_params=SummaryParams()):
    """
    Create model parameters

    Parameters
    ----------

    models : list Panda's data frames
        List of data frames for each model, containg sample values for
        multiple parameters (one parameter is one data frame column).
        Supply multiple data frames to compare parameter distributions.

    labels : list of str
        Names of the models in `models` list.

    extra_values : list of dict
        Additional values to be shown in the table:

        [
            {
                "mu": 2.3,
                "sigma": 3.3
            }
        ]

    type : CompareParametersType
        Format of values in the text table.

    param_names : list of str
        Names of parameters. Include all if None.

    Returns
    --------
    df: Panda's data frame
        Table in Panda's format
    txt : str
        Table in text format
    """

    if len(models) == 0:
        raise ValueError('Models list is empty')
        return

    if (len(models) + len(extra_values)) != len(labels):
        raise ValueError('Models list length is different from labels')
        return

    samples = models[0]
    column_names = list(samples)
    param_names = filter_param_names(column_names, param_names)

    df = pd.DataFrame(index=labels, columns=param_names)
    param_names_filtered = None

    for samples, label in zip(models, labels):
        column_names = list(samples)

        if param_names_filtered is None:
            param_names_filtered = filter_param_names(column_names, param_names)

        samples = samples[param_names_filtered]
        df_summary, _ = sample_summary(samples, params=summary_params)

        values = [
            format_parameter(df_summary.loc[name], type)
            for name in param_names_filtered
        ]

        df.loc[label] = values

    # Add extra values
    # ---------------

    extra_labels = labels[len(models):]

    for data, label in zip(extra_values, extra_labels):
        column_names = list(data.keys())
        values = []

        for name in param_names_filtered:
            text_value = ''

            if name in data:
                text_value = format_value(data[name], type)

            values.append(text_value)

        df.loc[label] = values

    table = tabulate(df, headers=param_names_filtered, tablefmt="pipe",
                     stralign="right")

    return df, table


def save_compare_parameters(
        models,
        labels,
        extra_values=[],
        type: CompareParametersType = CompareParametersType.TEXT,
        param_names=None,
        info_path=InfoPath(),
        summary_params=SummaryParams()):
    """
    Saves a text table that compares model parameters

    Parameters
    ----------

    models : list Panda's data frames
        List of data frames for each model, containg sample values for
        multiple parameters (one parameter is one data frame column).
        Supply multiple data frames to compare parameter distributions.

    labels : list of str
        Names of the models in `models` list.

    extra_values : list of dict
        Additional values to be shown in the table:

        [
            {
                "mu": 2.3,
                "sigma": 3.3
            }
        ]

    type : CompareParametersType
        Format of values in the text table.

    param_names : list of str
        Names of parameters. Include all if None.

    info_path : InfoPath
        Path information for creating summaries.
    """

    info_path.set_codefile()

    df, table = compare_parameters(models=models,
                                   labels=labels,
                                   extra_values=extra_values,
                                   type=type,
                                   param_names=param_names,
                                   summary_params=summary_params)

    info_path = InfoPath(**info_path.__dict__)
    info_path.base_name = info_path.base_name or "parameters_compared"
    info_path.extension = 'txt'
    path_to_txt = get_info_path(info_path)

    with open(path_to_txt, "w") as text_file:
        print(table, file=text_file)



from pytest import approx
import os
import shutil
import pandas as pd
import numpy as np
import pytest

from tarpan.cmdstanpy.psis import (
    psis, compare_psis, PsisData, PsisModelCompared, psis_compared_to_df,
    save_compare_psis_csv, save_compare_psis_txt,
    compare_psis_tree_plot, save_compare_psis_tree_plot,
    psis_pareto_k_plot, save_psis_pareto_k_plot,
    save_psis_pareto_k_plot_from_compared)

from tarpan.testutils.a03_cars.cars import get_fit

from tarpan.testutils.a04_height.height import get_fit1_intercept

from tarpan.testutils.a05_divorse.divorse import (
    get_fit1_divorse_age, get_fit2_divorse_marriage,
    get_fit3_divorse_age_marriage)


def test_psis():
    fit = get_fit1_divorse_age()

    result = psis(fit)

    assert result.psis == approx(126.26, rel=1e-4)
    assert result.psis_std_err == approx(12.981, rel=1e-4)
    assert result.penalty == approx(3.8938, rel=1e-4)

    assert len(result.psis_pointwise) == 50
    assert result.psis_pointwise[0] == approx(4.1495, rel=1e-4)
    assert result.psis_pointwise[49] == approx(1.9286, rel=1e-4)

    assert len(result.pareto_k) == 50
    assert result.pareto_k[0] == approx(0.093890, rel=1e-4)
    assert result.pareto_k[12] == approx(0.74203, rel=1e-4)
    assert result.pareto_k[49] == approx(0.13764, rel=1e-4)


def test_compare_psis():
    fit1_divorse_age = get_fit1_divorse_age()
    fit2_divorse_marriage = get_fit2_divorse_marriage()
    fit3_divorse_age_marriage = get_fit3_divorse_age_marriage()

    models = {
        "Divorse vs Age": fit1_divorse_age,
        "Divorse vs Marriage": fit2_divorse_marriage,
        "Divorse vs Age+Marriage": fit3_divorse_age_marriage
    }

    result = compare_psis(models=models)

    assert [model.name for model in result] == ['Divorse vs Age',
                                                'Divorse vs Age+Marriage',
                                                'Divorse vs Marriage']

    assert [round(model.psis_data.psis, 2) for model in result] == \
        [126.26, 127.07, 139.23]

    assert [round(model.psis_data.psis_std_err, 2) for model in result] == \
        [12.98, 12.38, 9.79]

    difference = [
        None if model.psis_difference_best is None
        else round(model.psis_difference_best, 2)
        for model in result
    ]

    assert difference == [None, 0.81, 12.98]

    std_err = [
        None if model.psis_difference_best_std_err is None
        else round(model.psis_difference_best_std_err, 2)
        for model in result
    ]

    assert std_err == [None, 1.14, 9.53]

    assert [round(model.psis_data.penalty, 1) for model in result] == \
        [3.9, 4.5, 2.9]

    actual_largest_k = [
        round(model.largest_pareto_k, 2)
        for model in result
    ]

    assert actual_largest_k == [0.74, 0.56, 0.39]

    actual_weight = [
        round(model.weight, 2)
        for model in result
    ]

    assert actual_weight == [0.57, 0.36, 0.08]


def test_compare_psis__model_with_different_data_points():
    cars_fit = get_fit()
    plant_fit = get_fit1_intercept()

    models = {
        "Cars": cars_fit,
        "Plants": plant_fit
    }

    with pytest.raises(AttributeError,
                       match=r"different number of data points"):
        compare_psis(models=models)


def test_psis_compared_to_df():
    compared = []

    for i in range(1, 4):
        psis = PsisData(
            psis=i,
            psis_pointwise=[i] * 3,
            psis_std_err=i * 1.1,
            lppd=i * 1.2,
            lppd_pointwise=[i * 1.2] * 3,
            penalty=i * 0.3,
            penalty_pointwise=[i * 0.3] * 3,
            pareto_k=[i * 1.5] * 3
        )

        compared_element = PsisModelCompared(
            name=f"Model {i}",
            psis_data=psis,
            psis_difference_best=i * 1.3,
            psis_difference_best_std_err=i * 1.4,
            largest_pareto_k=i * 1.6,
            weight=i * 1.7
        )

        compared.append(compared_element)

    result = psis_compared_to_df(compared=compared)

    assert len(result) == 3

    row = result.loc["Model 1"]
    assert row["PSIS"] == 1
    assert row["SE"] == 1.1
    assert row["dPSIS"] == 1.3
    assert row["dSE"] == 1.4
    assert row["pPSIS"] == 0.3
    assert row["MaxK"] == 1.6
    assert row["Weight"] == 1.7

    row = result.loc["Model 2"]
    assert row["PSIS"] == 2
    assert row["SE"] == 2.2
    assert row["dPSIS"] == 2.6
    assert row["dSE"] == 2.8
    assert row["pPSIS"] == 0.6
    assert row["MaxK"] == 3.2
    assert row["Weight"] == 3.4


def test_save_compare_psis_csv():
    fit1_divorse_age = get_fit1_divorse_age()
    fit2_divorse_marriage = get_fit2_divorse_marriage()
    fit3_divorse_age_marriage = get_fit3_divorse_age_marriage()

    models = {
        "Divorse vs Age": fit1_divorse_age,
        "Divorse vs Marriage": fit2_divorse_marriage,
        "Divorse vs Age+Marriage": fit3_divorse_age_marriage
    }

    outdir = "tarpan/cmdstanpy/model_info/psis_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_compare_psis_csv(models=models)

    assert os.path.isfile(os.path.join(outdir, "compare_psis.csv"))

    df = pd.read_csv(os.path.join(outdir, "compare_psis.csv"),
                     index_col="Name")

    assert len(df) == 3

    row = df.loc["Divorse vs Age"]
    assert row["PSIS"] == approx(126.25, rel=1e-3)
    assert row["SE"] == approx(12.981, rel=1e-3)
    assert np.isnan(row["dPSIS"])
    assert np.isnan(row["dSE"])
    assert row["pPSIS"] == approx(3.8938, rel=1e-3)
    assert row["MaxK"] == approx(0.74203, rel=1e-3)
    assert row["Weight"] == approx(0.5665, rel=1e-3)

    row = df.loc["Divorse vs Age+Marriage"]
    assert row["PSIS"] == approx(127.06, rel=1e-3)
    assert row["SE"] == approx(12.3822, rel=1e-3)
    assert row["dPSIS"] == approx(0.80989, rel=1e-3)
    assert row["dSE"] == approx(1.14009, rel=1e-3)
    assert row["pPSIS"] == approx(4.45447, rel=1e-3)
    assert row["MaxK"] == approx(0.56233, rel=1e-3)
    assert row["Weight"] == approx(0.3552, rel=1e-3)


def test_save_compare_psis_txt():
    fit1_divorse_age = get_fit1_divorse_age()
    fit2_divorse_marriage = get_fit2_divorse_marriage()
    fit3_divorse_age_marriage = get_fit3_divorse_age_marriage()

    models = {
        "Divorse vs Age": fit1_divorse_age,
        "Divorse vs Marriage": fit2_divorse_marriage,
        "Divorse vs Age+Marriage": fit3_divorse_age_marriage
    }

    outdir = "tarpan/cmdstanpy/model_info/psis_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_compare_psis_txt(models=models)

    assert os.path.isfile(os.path.join(outdir, "compare_psis.txt"))

    with open(os.path.join(outdir, "compare_psis.txt"), 'r') as file:
        data = file.read()
        assert "dPSIS" in data
        assert "Divorse vs Age" in data
        assert "12.98" in data


def test_compare_psis_tree_plot():
    fit1_divorse_age = get_fit1_divorse_age()
    fit2_divorse_marriage = get_fit2_divorse_marriage()
    fit3_divorse_age_marriage = get_fit3_divorse_age_marriage()

    models = {
        "Divorse vs Age": fit1_divorse_age,
        "Divorse vs Marriage": fit2_divorse_marriage,
        "Divorse vs Age+Marriage": fit3_divorse_age_marriage
    }

    fig, ax = compare_psis_tree_plot(models=models)

    assert ax.get_xlabel() == "PSIS (deviance)"


def test_save_compare_psis_tree_plot():
    fit1_divorse_age = get_fit1_divorse_age()
    fit2_divorse_marriage = get_fit2_divorse_marriage()
    fit3_divorse_age_marriage = get_fit3_divorse_age_marriage()

    models = {
        "Divorse vs Age": fit1_divorse_age,
        "Divorse vs Marriage": fit2_divorse_marriage,
        "Divorse vs Age+Marriage": fit3_divorse_age_marriage
    }

    outdir = "tarpan/cmdstanpy/model_info/psis_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_compare_psis_tree_plot(models=models)

    assert os.path.isfile(os.path.join(outdir, "compare_psis.pdf"))


def test_psis_pareto_k_plot():
    fit_divorse_age = get_fit1_divorse_age()

    fig, ax = psis_pareto_k_plot(fit=fit_divorse_age, name="Divorse vs age")

    assert ax.get_xlabel() == "Point index"
    assert ax.get_title() == "'Divorse vs age' Pareto K values"


def test_save_psis_pareto_k_plot():
    fit_divorse_age = get_fit1_divorse_age()

    outdir = "tarpan/cmdstanpy/model_info/psis_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_psis_pareto_k_plot(fit=fit_divorse_age, name="Divorse vs age")

    assert os.path.isfile(os.path.join(outdir, "pareto_k_divorse_vs_age.pdf"))


def test_save_psis_pareto_k_plot_from_compared():
    outdir = "tarpan/cmdstanpy/model_info/psis_test"

    compared = []

    for i in range(1, 4):
        psis = PsisData(
            psis=i,
            psis_pointwise=[i] * 3,
            psis_std_err=i * 1.1,
            lppd=i * 1.2,
            lppd_pointwise=[i * 1.2] * 3,
            penalty=i * 0.3,
            penalty_pointwise=[i * 0.3] * 3,
            pareto_k=[i * 1.5] * 3
        )

        compared_element = PsisModelCompared(
            name=f"Model {i}",
            psis_data=psis,
            psis_difference_best=i * 1.3,
            psis_difference_best_std_err=i * 1.4,
            largest_pareto_k=i * 1.6,
            weight=i * 1.7
        )

        compared.append(compared_element)

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_psis_pareto_k_plot_from_compared(compared=compared)

    assert os.path.isfile(os.path.join(outdir, "pareto_k_model_1.pdf"))
    assert os.path.isfile(os.path.join(outdir, "pareto_k_model_2.pdf"))
    assert os.path.isfile(os.path.join(outdir, "pareto_k_model_3.pdf"))



import shutil
import os
import pandas as pd
from tarpan.shared.histogram import save_histogram


def test_save_tree_plot():
    data = {
        "x": [1, 2, 3, 4, 5, 6, 2, 3, 4, 4, 1, 3, 4, 5, 6, 2],
        "y": [-1, -2, -3, -4, -5, -6, -2, -2, -1, -1, -1, -3, -3, -2, -1, -4]
    }

    df = pd.DataFrame(data)

    outdir = "tarpan/shared/model_info/histogram_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_histogram(df, param_names=["x", "y"])

    assert os.path.isfile(os.path.join(outdir, "histogram_01.pdf"))



import shutil
import os
import pandas as pd
from tarpan.shared.summary import save_summary


def test_save_tree_plot():
    data = {
        "x": [1, 2, 3, 4, 5, 6],
        "y": [-1, -2, -3, -4, -5, -6]
    }

    df = pd.DataFrame(data)

    outdir = "tarpan/shared/model_info/summary_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    result = save_summary(df, param_names=["x"])

    assert os.path.isfile(os.path.join(outdir, "summary.txt"))
    assert os.path.isfile(os.path.join(outdir, "summary.csv"))

    assert result["df"].shape == (1, 9)
    assert "1.87" in result["table"]
    assert os.path.join(outdir, "summary.txt") in result["path_txt"]
    assert os.path.join(outdir, "summary.csv") in result["path_csv"]



import os
import shutil
from tarpan.cmdstanpy.pair_plot import save_pair_plot
from tarpan.testutils.a01_eight_schools.eight_schools import get_fit


def test_save_pair_plot():
    fit = get_fit()

    outdir = "tarpan/cmdstanpy/model_info/pair_plot_test"

    if os.path.isdir(outdir):
        shutil.rmtree(outdir)

    save_pair_plot(fit, param_names=["mu", "tau", 'eta.1'])

    assert os.path.isfile(os.path.join(outdir, "pair_plot.pdf"))
